class
	AST_VARIABLE_DECLARATION

inherit
	AST_STRUCT_FIELD
		redefine
			make_copy
		end
	AST_EIFFEL_FEATURE
		redefine
			make_copy
		end

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_type: AST_TYPE; a_identifier: STRING; a_initializer: AST_EXPRESSION)
		require
			a_type /= Void and then a_type.parent = Void
			a_identifier /= Void
			a_initializer /= Void implies a_initializer.parent = Void
		do
			storage_class_specifiers := create {TWO_WAY_LIST [INTEGER]}.make
			a_type.parent := Current
			type := a_type
			identifier := a_identifier
			orig_c_identifier := identifier.twin
			if a_initializer /= Void then
				a_initializer.parent := Current
				initializer := a_initializer
			end
		ensure
			type = a_type
			identifier = a_identifier
			initializer = a_initializer
		end

	make_from_antlr (node: ANTLR_AST_NODE; a_parent: AST_NODE; a_source_info: SOURCE_INFO)
		require
			node /= Void and then node.type = Ast_variable_declaration_or_type
		local
			l_h: AST_VARIABLE_DECLARATION_OR_TYPE_HELPER
		do
			parent := a_parent

			l_h := create {AST_VARIABLE_DECLARATION_OR_TYPE_HELPER}.make_from_antlr (node)
			identifier := l_h.identifier
			orig_c_identifier := identifier.twin
			storage_class_specifiers := l_h.storage_class_specifiers
			type := l_h.type
			type.parent := Current
			gcc_attribute_specifier := l_h.gcc_attribute_specifier
			if gcc_attribute_specifier /= Void then
				gcc_attribute_specifier.parent := Current
			end
			initializer := l_h.initializer
			if initializer /= Void then
				initializer.parent := Current
			end
			source_info := a_source_info
		end

feature {ANY}

	type: AST_TYPE assign set_type

	initializer: AST_EXPRESSION

	set_type (a_type: AST_TYPE)
		require
			a_type /= Void and then a_type.parent = Void
		do
			type.parent := Void
			type := a_type
			type.parent := Current
		end

feature {ANY} -- C2Eif data

	changed_to_array: BOOLEAN assign set_changed_to_array

	set_changed_to_array (a_changed_to_array: BOOLEAN)
		do
			changed_to_array := a_changed_to_array
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_variable_declaration (Current)
			if visit_children then
				visit_node (type, v)
				visit_node (gcc_attribute_specifier, v)
				visit_node (initializer, v)
			end
			v.end_visit_variable_declaration (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (a_other: like Current)
		do
			Precursor {AST_STRUCT_FIELD} (a_other)
			Precursor {AST_EIFFEL_FEATURE} (a_other)
			type := a_other.type.ast_twin
			type.parent := Current
			if a_other.initializer /= Void then
				initializer := a_other.initializer.ast_twin
				initializer.parent := Current
			end
			changed_to_array := a_other.changed_to_array
		end

invariant
	type /= Void
	initializer /= Void implies (parent = Void or attached {AST_TRANSLATION_UNIT} parent or attached {AST_ENUM_DECLARATION} parent)

end
