deferred class
	AST_EXPRESSION

inherit
	AST_NODE
	BINDINGS_STATUS

feature {ANY}

	parentheses_parent: AST_EXPRESSION
			-- Returns either the node itself or the outermost surrounding parentheses node.
			-- For `a' in `a + 1', `a' is returned.
			-- For `a' in `(*((a))) + 1', `((a))' is returned.
		do
			if attached {AST_PARENTHESIZED_EXPRESSION} parent as pe then
				Result := pe.parentheses_parent
			else
				Result := Current
			end
		ensure
			Result = Current or attached {AST_PARENTHESIZED_EXPRESSION} Result
		end

	non_parentheses_parent: AST_NODE
		require
			parent /= Void
		do
			from
				Result := parent
			until
				not attached {AST_PARENTHESIZED_EXPRESSION} Result
			loop
				Result := Result.parent
			end
		ensure
			Result /= Void and then not attached {AST_PARENTHESIZED_EXPRESSION} Result
		end

	non_parentheses_expression: AST_EXPRESSION
			-- Removes parentheses
		do
			from
				Result := Current
			until
				not attached {AST_PARENTHESIZED_EXPRESSION} Result as tmp
			loop
				Result := tmp.expression
			end
		ensure
			Result /= Void
		end

	non_cast_expression: AST_EXPRESSION
			-- Removes casts and parentheses
		do
			from
				Result := non_parentheses_expression
			until
				not attached {AST_CAST_EXPRESSION} Result as ce
			loop
				Result := ce.expression.non_parentheses_expression
			end
		ensure
			Result /= Void
		end

feature {ANY} -- type binding

	type_binding: AST_TYPE
		deferred
		end

	is_orig_c_expression: BOOLEAN assign set_is_orig_c_expression

	set_is_orig_c_expression (a_is_orig_c_expression: BOOLEAN)
		do
			is_orig_c_expression := a_is_orig_c_expression
		end

	is_in_orig_c_expression: BOOLEAN
		do
			if is_orig_c_expression then
				Result := True
			elseif attached {AST_EXPRESSION} parent as tmp then
				Result := tmp.is_in_orig_c_expression
			end
		end

invariant
	(Bindings_active.item and not is_in_orig_c_expression) implies type_binding /= Void

end
