class
	AST_UNARY_EXPRESSION

inherit
	AST_EXPRESSION
	AST_UTILITIES

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_type: INTEGER; a_expression: AST_EXPRESSION)
		require
			is_valid_type (a_type)
			a_expression /= Void and then a_expression.parent = Void
		do
			type := a_type
			expression := a_expression
			expression.parent := Current
		end

	make_from_antlr (node: ANTLR_AST_NODE; init_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_unary_expression
		do
			parent := init_parent
			type := node.children.first.type
			expression := process_expression (node.children.i_th (2), Current)
		end

feature {ANY}

	type: INTEGER

	expression: AST_EXPRESSION

	type_binding: AST_TYPE
		once ("OBJECT")
			inspect
				type
			when C_op_increment, C_op_decrement, C_op_plus, C_op_minus, C_op_neg then
				Result := expression.type_binding
			when C_op_bitand then
				Result := create {AST_POINTER_TYPE}.make (expression.type_binding.ast_twin)
			when C_op_mult then
				Result := deref (expression.type_binding)
			when C_op_not then
				-- becomes `not' in Eiffel, so we say it's boolean
				Result := create {AST_BASIC_TYPE}.make (Ast_bt_bool)
			end
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_unary_expression (Current)
			if visit_children then
				visit_node (expression, v)
			end
			v.end_visit_unary_expression (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			type := other.type
			expression := other.expression.ast_twin
			expression.parent := Current
		end

feature {NONE}

	is_valid_type (a_type: INTEGER): BOOLEAN
		do
			inspect
				a_type
			when C_op_increment, C_op_decrement, C_op_bitand, C_op_mult, C_op_plus, C_op_minus, C_op_neg, C_op_not then
				Result := True
			else
				Result := False
			end
		end

invariant
	is_valid_type (type)
	expression /= Void

end
