class
	AST_ASSEMBLER_STATEMENT

inherit
	AST_STATEMENT
		redefine
			is_equal
		end

create
	make_from_antlr,
	make_copy

feature {NONE}

	make_from_antlr (a_node: ANTLR_AST_NODE; a_parent: AST_NODE)
		require
			a_node /= Void and then a_node.type = Ast_assembler_statement
		local
			l_tmp: STRING_8
		do
			parent := a_parent

			a_node.children.start
			template := create {TWO_WAY_LIST [STRING]}.make
			across a_node.children.item.children as c loop
				template.extend (c.item.token_text)
			end
			a_node.children.forth

			output_operands := create {TWO_WAY_LIST [AST_ASM_OPERAND]}.make
			if not a_node.children.after then
				a_node.children.forth -- skip ':'
				if
					not a_node.children.after and then
					a_node.children.item.type /= C_op_colon
				then
					across a_node.children.item.children as c loop
						l_tmp := c.item.token_text
						c.forth
						output_operands.extend (create {AST_ASM_OPERAND}.make (l_tmp, process_expression (c.item, Void)))
						output_operands.last.parent := Current
					end
					a_node.children.forth
				end
			end

			input_operands := create {TWO_WAY_LIST [AST_ASM_OPERAND]}.make
			if not a_node.children.after then
				a_node.children.forth -- skip ':'
				if
					not a_node.children.after and then
					a_node.children.item.type /= C_op_colon
				then
					across a_node.children.item.children as c loop
						l_tmp := c.item.token_text
						c.forth
						input_operands.extend (create {AST_ASM_OPERAND}.make (l_tmp, process_expression (c.item, Void)))
						input_operands.last.parent := Current
					end
					a_node.children.forth
				end
			end

			clobbered_regs := create {TWO_WAY_LIST [STRING]}.make
			if not a_node.children.after then
				a_node.children.forth -- skip ':'
				if not a_node.children.after then
					across a_node.children.item.children as c loop
						clobbered_regs.extend (c.item.token_text)
					end
				end
			end
		end

feature {ANY}

	template: LIST [STRING]

	output_operands: LIST [AST_ASM_OPERAND]

	input_operands: LIST [AST_ASM_OPERAND]

	clobbered_regs: LIST [STRING]

feature {ANY}

	is_equal (a_other: like Current): BOOLEAN
		do
			-- TODO: remove `accessed_variables', merge equal assembler statements

--			Result :=
--				template.count = a_other.template.count and
--				output_operands.count = a_other.output_operands.count and
--				input_operands.count = a_other.input_operands.count and
--				clobbered_regs.count = a_other.clobbered_regs.count

--			template.start
--			across
--				a_other.template as c
--			until
--				not Result
--			loop
--				Result := c.item.is_equal (template.item)
--				template.forth
--			end

--			output_operands.start
--			across
--				a_other.output_operands as c
--			until
--				not Result
--			loop
--				Result := c.item.constraint.is_equal (output_operands.item.constraint)
--				output_operands.forth
--			end

--			input_operands.start
--			across
--				a_other.input_operands as c
--			until
--				not Result
--			loop
--				Result := c.item.constraint.is_equal (input_operands.item.constraint)
--				input_operands.forth
--			end

--			clobbered_regs.start
--			across
--				a_other.clobbered_regs as c
--			until
--				not Result
--			loop
--				Result := c.item.is_equal (clobbered_regs.item)
--				clobbered_regs.forth
--			end
		end

	accessed_variables: LIST [AST_VARIABLE_EXPRESSION]
		local
			agent_visitor: AST_AGENT_VISITOR
		once ("OBJECT")
			Result := create {TWO_WAY_LIST [AST_VARIABLE_EXPRESSION]}.make

			create agent_visitor
			agent_visitor.agents.end_visit_variable_expression := agent (n: AST_VARIABLE_EXPRESSION; av: LIST [AST_VARIABLE_EXPRESSION])
				do
					if across av as c all not c.item.identifier.is_equal (n.identifier) end then
						av.extend (n)
					end
				end (?, Result)

			across output_operands as c loop
				c.item.expression.accept (agent_visitor)
			end
			across input_operands as c loop
				c.item.expression.accept (agent_visitor)
			end
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_assembler_statement (Current)
			if visit_children then
				visit_list (output_operands, v)
				visit_list (input_operands, v)
			end
			v.end_visit_assembler_statement (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			template := other.template.twin
			output_operands := create {TWO_WAY_LIST [AST_ASM_OPERAND]}.make
			across other.output_operands as c loop
				output_operands.extend (c.item.ast_twin)
				output_operands.last.parent := Current
			end
			input_operands := create {TWO_WAY_LIST [AST_ASM_OPERAND]}.make
			across other.input_operands as c loop
				input_operands.extend (c.item.ast_twin)
				input_operands.last.parent := Current
			end
			clobbered_regs := other.clobbered_regs.twin
		end

invariant
	template /= Void and then not template.is_empty
	output_operands /= Void
	input_operands /= Void
	clobbered_regs /= Void

end
