class
	AST_STRUCT_TYPE

inherit
	AST_TYPE
		redefine
			make_copy
		end
	BINDINGS_STATUS

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_identifier: STRING; a_is_union: BOOLEAN; a_declaration_binding: AST_STRUCT_DECLARATION)
		require
			a_identifier /= Void
			Bindings_active.item implies a_declaration_binding /= Void
		do
			identifier := a_identifier
			is_union := a_is_union
			declaration_binding := a_declaration_binding
		end

	make_from_antlr (node: ANTLR_AST_NODE; init_parent: AST_NODE)
		require
			node /= Void and then (node.type = C_kw_struct or node.type = C_kw_union)
		do
			parent := init_parent
			identifier := node.children.first.token_text
			is_union := node.type = C_kw_union
		end

feature {ANY}

	identifier: STRING

	is_union: BOOLEAN

	declaration_binding: AST_STRUCT_DECLARATION assign set_declaration_binding

	set_declaration_binding (a_declaration_binding: AST_STRUCT_DECLARATION)
		require
			a_declaration_binding /= Void
		do
			declaration_binding := a_declaration_binding
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_struct_type (Current)
			if visit_children then
				-- no children
			end
			v.end_visit_struct_type (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			Precursor (other)
			identifier := other.identifier.twin
			is_union := other.is_union
			declaration_binding := other.declaration_binding
		end

feature {ANY} -- AST_TYPE

	is_equal_type (a_type: AST_TYPE): BOOLEAN
		do
			if attached {AST_STRUCT_TYPE} a_type as other_type then
				Result := is_union = other_type.is_union and identifier.is_equal (other_type.identifier)
			end
		end


invariant
	identifier /= Void and then not identifier.is_empty
	-- Bindings_active.item implies declaration_binding /= Void -> not if in SOURCE_LOADER.orig_declarations

end
