class
	CFG_CREATOR

inherit
	TRANSFORMATION_STEP
	AST_VISITOR

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			create breaks.make

			across
				transformer.tree.declarations as c
			invariant
				breaks.is_empty
			loop
				if
					attached {AST_FUNCTION_DECLARATION} c.item as fd and then
					fd.is_regular
				then
					create start_block
					cur_block := start_block
					jumps := create {TWO_WAY_LIST [TUPLE [CFG_BASIC_BLOCK, NATURAL_32]]}.make
					create jump_targets.make (16)

					fd.statements.accept (Current)
					across jumps as c2 loop
						(create {CFG_EDGE}.make (c2.item.block, jump_targets[c2.item.state], {CFG_EDGE}.Label_no_label)).do_nothing
					end
					remove_empty_blocks
					merge_blocks

					fd.cfg := start_block
				end
			end

			start_block := Void
			cur_block := Void
			jumps := Void
			jump_targets := Void
			breaks := Void

			Result := True
		end

	description: STRING
		do
			Result := "Creating the control flow graphs"
		end

feature {NONE}

	start_block, cur_block: CFG_BASIC_BLOCK

	jumps: LIST [TUPLE [block: CFG_BASIC_BLOCK; state: NATURAL_32]]

	jump_targets: HASH_TABLE [CFG_BASIC_BLOCK, NATURAL_32]

	breaks: LINKED_STACK [TUPLE [blocks_with_break: HASH_SET [CFG_BASIC_BLOCK]; end_jump_state: NATURAL_32_REF]]

feature {NONE}

	remove_empty_blocks
		do
			if not start_block.out_edges.is_empty then
				remove_empty_blocks_rec (start_block, create {HASH_SET [CFG_BASIC_BLOCK]}.make (16))
			end
		end

	remove_empty_blocks_rec (a_block: CFG_BASIC_BLOCK; a_visited: HASH_SET [CFG_BASIC_BLOCK])
		require
			a_block /= Void
			a_visited /= Void
		do
			if not a_visited.has (a_block) then
				a_visited.extend (a_block)

				across a_block.out_edges.linear_representation as c loop
					remove_empty_blocks_rec (c.item.to_block, a_visited)
				end

				if a_block.nodes.is_empty then
					if not a_block.in_edges.is_empty and not a_block.out_edges.is_empty then
						if
							a_block.out_edges.count = 1 and then
							attached a_block.out_edges.linear_representation.first.to_block as out_block
						then
							out_block.add_in_edges (a_block.in_edges)
							a_block.remove_in_edges
							a_block.remove_out_edges
						else
							(create {EXCEPTIONS}).raise ("?!")
						end
					elseif not a_block.out_edges.is_empty then
						if a_block = start_block and a_block.out_edges.count = 1 then
							start_block := a_block.out_edges.linear_representation.first.to_block
							a_block.get_out_edge (start_block).remove
						else
							(create {EXCEPTIONS}).raise ("?!")
						end
					elseif not a_block.in_edges.is_empty then
						if
							a_block.in_edges.count = 1 and
							a_block.in_edges.linear_representation.first.from_block.out_edges.count = 1
						then
							a_block.remove_in_edges
						end
					else
						(create {EXCEPTIONS}).raise ("?!")
					end
				end
			end
		end

	merge_blocks
		do
			merge_blocks_rec (start_block, create {HASH_SET [CFG_BASIC_BLOCK]}.make (16))
		end

	merge_blocks_rec (a_block: CFG_BASIC_BLOCK; a_visited: HASH_SET [CFG_BASIC_BLOCK])
		require
			a_block /= Void
			a_visited /= Void
		do
			if not a_visited.has (a_block) then
				a_visited.extend (a_block)

				if
					a_block.out_edges.count = 1 and then
					attached a_block.out_edges.linear_representation.first.to_block as out_block and then
					out_block.in_edges.count = 1 and then
					out_block /= start_block -- second "in edge"
				then
					across out_block.nodes as c loop
						a_block.nodes.extend (c.item)
					end
					a_block.remove_out_edges
					a_block.add_out_edges (out_block.out_edges)
					out_block.remove_out_edges
				end

				across a_block.out_edges as c loop
					merge_blocks_rec (c.item.to_block, a_visited)
				end
			end
		end

feature {NONE}

	visit_compound_statement (a_node: AST_COMPOUND_STATEMENT): BOOLEAN
		local
			l_switch: LIST [AST_EIFFEL_EXPRESSION]
		do
			across a_node.statements as c loop
				if
					attached {AST_EXPRESSION_STATEMENT} c.item as es and then
					attached {AST_EIFFEL_EXPRESSION} es.expression as ee and then
					ee.is_string (1, "inspect ")
				then
					l_switch := create {TWO_WAY_LIST [AST_EIFFEL_EXPRESSION]}.make
					l_switch.extend (ee)
				elseif l_switch /= Void then
					if
						attached {AST_EXPRESSION_STATEMENT} c.item as es and then
						attached {AST_EIFFEL_EXPRESSION} es.expression as ee
					then
						l_switch.extend (ee)
						if ee.is_string (1, "end") then
							handle_switch (l_switch)
							l_switch := Void
						end
					else
						(create {EXCEPTIONS}).raise ("?!")
					end
				else
					c.item.accept (Current)
				end
			end
		end

	visit_expression_statement (a_node: AST_EXPRESSION_STATEMENT): BOOLEAN
		local
			l_is_return, l_is_break: BOOLEAN
		do
			if
				attached {AST_EIFFEL_EXPRESSION} a_node.expression as ee and then
				ee.mixed.count = 1 and then
				attached {STRING_8} ee.mixed.first as s
			then
				    if s.is_equal ("l_return := True")   then   l_is_return := True
				elseif s.is_equal ("l_break := True")    then    l_is_break := True
				--TODO: elseif s.is_equal ("l_continue := True") then l_is_continue := True
				end
			end

			if l_is_return or l_is_break then
				if l_is_break then
					breaks.item.blocks_with_break.extend (cur_block)
				end

				-- stop the cfg
				cur_block.remove_out_edges
				cur_block := Void

				-- add a "dummy" cur_block for potential dead code
				create cur_block
			else
				cur_block.nodes.extend (a_node)
			end
		end

	visit_if_statement (a_node: AST_IF_STATEMENT): BOOLEAN
		local
			l_is_loop: BOOLEAN
			l_head, l_then, l_else, l_tail: CFG_BASIC_BLOCK
		do
			if
				a_node.else_statements = Void and then
				attached {AST_UNARY_EXPRESSION} a_node.condition as ue and then
				ue.type = C_op_not and then
				attached {AST_EIFFEL_EXPRESSION} ue.expression as ee and then
				ee.mixed.count = 1 and then
				ee.is_string (1, "l_return")
			then
				-- `if(!l_return) { ... }'
				-- we stop the cfg after `l_return := True'; just use the `...'
				a_node.then_statements.accept (Current)
			else
				if
					a_node.else_statements = Void and then
					(
						a_node.condition.c_code.has_substring ("(l_return)") or
						a_node.condition.c_code.has_substring ("(l_break)") or
						a_node.condition.c_code.has_substring ("(l_return or l_break)")
					) and then
					a_node.then_statements.statements.count = 1 and then
					attached {AST_EIFFEL_JUMP} a_node.then_statements.statements.first as ej
				then
					l_is_loop := True
					breaks.extend ([create {HASH_SET [CFG_BASIC_BLOCK]}.make (16), ej.index])
				end

				l_head := cur_block
				create l_then
				if a_node.else_statements /= Void then
					create l_else
				end
				create l_tail

				l_tail.add_out_edges (l_head.out_edges)
				l_head.remove_out_edges

				if
					not (
							l_is_loop and
							(
								a_node.condition.c_code.is_equal ("(l_return)") or
								a_node.condition.c_code.is_equal ("(l_break)") or
								a_node.condition.c_code.is_equal ("(l_return or l_break)")
							)
						)
				then
					(create {CFG_EDGE}.make (l_head, l_then, {CFG_EDGE}.Label_true)).do_nothing
					l_head.nodes.extend (a_node.condition)
				end
				if l_else /= Void then
					(create {CFG_EDGE}.make (l_head, l_else, {CFG_EDGE}.Label_false)).do_nothing;
					(create {CFG_EDGE}.make (l_else, l_tail, {CFG_EDGE}.Label_no_label)).do_nothing
				else
					(create {CFG_EDGE}.make (l_head, l_tail, {CFG_EDGE}.Label_false)).do_nothing
				end
				(create {CFG_EDGE}.make (l_then, l_tail, {CFG_EDGE}.Label_no_label)).do_nothing

				cur_block := l_then
				a_node.then_statements.accept (Current)
				if l_else /= Void then
					cur_block := l_else
					a_node.else_statements.accept (Current)
				end
				cur_block := l_tail
			end
		end

	visit_while_statement (a_node: AST_WHILE_STATEMENT): BOOLEAN
		local
			l_head, l_if, l_then, l_tail: CFG_BASIC_BLOCK
		do
			breaks.extend ([create {HASH_SET [CFG_BASIC_BLOCK]}.make (16), Void])

			l_head := cur_block
			create l_if
			create l_then
			create l_tail

			l_tail.add_out_edges (l_head.out_edges)
			l_head.remove_out_edges;

			(create {CFG_EDGE}.make (l_head, l_if, {CFG_EDGE}.Label_no_label)).do_nothing;
			(create {CFG_EDGE}.make (l_if, l_then, {CFG_EDGE}.Label_no_label)).do_nothing
			if
				not (
						a_node.condition.c_code.is_equal ("!(l_return)") or
						a_node.condition.c_code.is_equal ("!(l_break)") or
						a_node.condition.c_code.is_equal ("!(l_return or l_break)")
					)
			then
				(create {CFG_EDGE}.make (l_if, l_tail, {CFG_EDGE}.Label_no_label)).do_nothing
				l_if.nodes.extend (a_node.condition)
			end
			(create {CFG_EDGE}.make (l_then, l_if, {CFG_EDGE}.Label_no_label)).do_nothing

			cur_block := l_then
			a_node.statements.accept (Current)
			cur_block := l_tail

			across breaks.item.blocks_with_break as c loop
				(create {CFG_EDGE}.make (c.item, cur_block, {CFG_EDGE}.Label_no_label)).do_nothing
			end
			breaks.remove
		end

	visit_eiffel_jump_target (a_node: AST_EIFFEL_JUMP_TARGET): BOOLEAN
		local
			l_new_cur_block: CFG_BASIC_BLOCK
		do
			create l_new_cur_block
			if cur_block /= Void then
				l_new_cur_block.add_out_edges (cur_block.out_edges)
				cur_block.remove_out_edges;
				(create {CFG_EDGE}.make (cur_block, l_new_cur_block, {CFG_EDGE}.Label_no_label)).do_nothing
			end
			cur_block := l_new_cur_block

			jump_targets.extend (l_new_cur_block, a_node.index)

			if
				not breaks.is_empty and then
				attached breaks.item.end_jump_state as index and then
				index.item = a_node.index.item
			then
				across breaks.item.blocks_with_break as c loop
					(create {CFG_EDGE}.make (c.item, cur_block, {CFG_EDGE}.Label_no_label)).do_nothing
				end
				breaks.remove
			end
		end

	visit_eiffel_jump (a_node: AST_EIFFEL_JUMP): BOOLEAN
		do
			if attached a_node.index as index_ref then
				jumps.extend ([cur_block, index_ref.item])
				cur_block.remove_out_edges
				cur_block := Void
			else
				cur_block.remove_out_edges
				cur_block := Void
			end
		end

	handle_switch (a_switch: LIST [AST_EIFFEL_EXPRESSION])
		require
			a_switch /= Void and then a_switch.count >= 2
			a_switch.first.is_string (1, "inspect ")
			a_switch.last.is_string (1, "end")
		local
			l_s: STRING_8
			l_jump: AST_EIFFEL_JUMP
		do
			across a_switch as c loop
				l_s ?= c.item.mixed.first
				if l_s.is_equal ("inspect ") then
					cur_block.nodes.extend ({AST_NODE} #? c.item.mixed.i_th (2))
					cur_block.remove_out_edges
				elseif
					l_s.starts_with ("when ") or
					l_s.is_equal ("else ")
				then
					l_jump ?= c.item.mixed.i_th (2)
					-- Don't duplicate jumps (successive switch cases point to the same block):
					if
						jumps.is_empty or else
						not across jumps as c2 some c2.item.block = cur_block and c2.item.state = l_jump.index.item end
					then
						jumps.extend ([cur_block, l_jump.index.item])
					end
				elseif c.item.is_string (1, "end") then
					-- nothing to do
				else
					(create {EXCEPTIONS}).raise ("%"" + l_s + "%"?!")
				end
			end
			cur_block := Void
		end

feature {NONE}

	visit_function_declaration (node: AST_FUNCTION_DECLARATION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_function_declaration") end
	visit_array_type (node: AST_ARRAY_TYPE): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_array_type") end
	visit_assembler_statement (node: AST_ASSEMBLER_STATEMENT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_assembler_statement") end
	visit_asm_operand (node: AST_ASM_OPERAND): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_asm_operand") end
	visit_attribute (node: AST_GCC_ATTRIBUTE): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_attribute") end
	visit_attribute_specifier (node: AST_GCC_ATTRIBUTE_SPECIFIER): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_attribute_specifier") end
	visit_basic_type (node: AST_BASIC_TYPE): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_basic_type") end
	visit_bit_field (node: AST_BITFIELD): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_bit_field") end
	visit_builtin_va_arg_call (node: AST_BUILTIN_VA_ARG_CALL): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_builtin_va_arg_call") end
	visit_alignof_call (node: AST_ALIGNOF_CALL): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_alignof_call") end
	visit_initializer (node: AST_INITIALIZER): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_initializer") end
	visit_assignment_expression (node: AST_ASSIGNMENT_EXPRESSION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_assignment_expression") end
	visit_boolean_constant (node: AST_BOOLEAN_CONSTANT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_boolean_constant") end
	visit_cast_expression (node: AST_CAST_EXPRESSION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_cast_expression") end
	visit_character_constant (node: AST_CHARACTER_CONSTANT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_character_constant") end
	visit_decimal_constant (node: AST_DECIMAL_CONSTANT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_decimal_constant") end
	visit_eiffel_expression (node: AST_EIFFEL_EXPRESSION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_eiffel_expression") end
	visit_eiffel_string_type (node: AST_EIFFEL_STRING_TYPE): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_eiffel_string_type") end
	visit_eiffel_type (node: AST_EIFFEL_TYPE): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_eiffel_type") end
	visit_enum_declaration (node: AST_ENUM_DECLARATION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_enum_declaration") end
	visit_enum_type (node: AST_ENUM_TYPE): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_enum_type") end
	visit_floating_point_constant (node: AST_FLOATING_POINT_CONSTANT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_floating_point_constant") end
	visit_function_call (node: AST_FUNCTION_CALL): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_function_call") end
	visit_function_type (node: AST_FUNCTION_TYPE): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_function_type") end
	visit_generic_expression (node: AST_GENERIC_EXPRESSION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_generic_expression") end
	visit_hex_constant (node: AST_HEX_CONSTANT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_hex_constant") end
	visit_octal_constant (node: AST_OCTAL_CONSTANT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_octal_constant") end
	visit_parameter_list (node: AST_PARAMETER_LIST): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_parameter_list") end
	visit_parenthesized_expression (node: AST_PARENTHESIZED_EXPRESSION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_parenthesized_expression") end
	visit_pointer_type (node: AST_POINTER_TYPE): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_pointer_type") end
	visit_postfix_expression (node: AST_POSTFIX_EXPRESSION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_postfix_expression") end
	visit_sizeof_expression (node: AST_SIZEOF_EXPRESSION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_sizeof_expression") end
	visit_string_constant (node: AST_STRING_CONSTANT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_string_constant") end
	visit_struct_declaration (node: AST_STRUCT_DECLARATION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_struct_declaration") end
	visit_struct_type (node: AST_STRUCT_TYPE): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_struct_type") end
	visit_switch_case (node: AST_SWITCH_CASE): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_switch_case") end
	visit_switch_default (node: AST_SWITCH_DEFAULT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_switch_default") end
	visit_switch_statement (node: AST_SWITCH_STATEMENT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_switch_statement") end
	visit_translation_unit (node: AST_TRANSLATION_UNIT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_translation_unit") end
	visit_typedef_declaration (node: AST_TYPEDEF_DECLARATION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_typedef_declaration") end
	visit_typedef_type (node: AST_TYPEDEF_TYPE): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_typedef_type") end
	visit_unary_expression (node: AST_UNARY_EXPRESSION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_unary_expression") end
	visit_variable_declaration (node: AST_VARIABLE_DECLARATION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_variable_declaration") end
	visit_variable_expression (node: AST_VARIABLE_EXPRESSION): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_variable_expression") end
	visit_labeled_statement (node: AST_LABELED_STATEMENT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_variable_expression") end
	visit_jump_statement (node: AST_JUMP_STATEMENT): BOOLEAN do (create {EXCEPTIONS}).raise ("visit_variable_expression") end

feature {NONE}

	pre_visit (node: AST_NODE) do end
	post_visit (node: AST_NODE) do end

feature {NONE}

	end_visit_array_type (node: AST_ARRAY_TYPE) do end
	end_visit_assembler_statement (node: AST_ASSEMBLER_STATEMENT) do end
	end_visit_asm_operand (node: AST_ASM_OPERAND) do end
	end_visit_attribute (node: AST_GCC_ATTRIBUTE) do end
	end_visit_attribute_specifier (node: AST_GCC_ATTRIBUTE_SPECIFIER) do end
	end_visit_basic_type (node: AST_BASIC_TYPE) do end
	end_visit_bit_field (node: AST_BITFIELD) do end
	end_visit_builtin_va_arg_call (node: AST_BUILTIN_VA_ARG_CALL) do end
	end_visit_alignof_call (node: AST_ALIGNOF_CALL) do end
	end_visit_initializer (node: AST_INITIALIZER) do end
	end_visit_assignment_expression (node: AST_ASSIGNMENT_EXPRESSION) do end
	end_visit_boolean_constant (node: AST_BOOLEAN_CONSTANT) do end
	end_visit_cast_expression (node: AST_CAST_EXPRESSION) do end
	end_visit_character_constant (node: AST_CHARACTER_CONSTANT) do end
	end_visit_compound_statement (node: AST_COMPOUND_STATEMENT) do end
	end_visit_decimal_constant (node: AST_DECIMAL_CONSTANT) do end
	end_visit_eiffel_expression (node: AST_EIFFEL_EXPRESSION) do end
	end_visit_eiffel_jump (node: AST_EIFFEL_JUMP) do end
	end_visit_eiffel_jump_target (node: AST_EIFFEL_JUMP_TARGET) do end
	end_visit_eiffel_string_type (node: AST_EIFFEL_STRING_TYPE) do end
	end_visit_eiffel_type (node: AST_EIFFEL_TYPE) do end
	end_visit_enum_declaration (node: AST_ENUM_DECLARATION) do end
	end_visit_enum_type (node: AST_ENUM_TYPE) do end
	end_visit_expression_statement (node: AST_EXPRESSION_STATEMENT) do end
	end_visit_floating_point_constant (node: AST_FLOATING_POINT_CONSTANT) do end
	end_visit_function_call (node: AST_FUNCTION_CALL) do end
	end_visit_function_declaration (node: AST_FUNCTION_DECLARATION) do end
	end_visit_function_type (node: AST_FUNCTION_TYPE) do end
	end_visit_generic_expression (node: AST_GENERIC_EXPRESSION) do end
	end_visit_hex_constant (node: AST_HEX_CONSTANT) do end
	end_visit_if_statement (node: AST_IF_STATEMENT) do end
	end_visit_jump_statement (node: AST_JUMP_STATEMENT) do end
	end_visit_labeled_statement (node: AST_LABELED_STATEMENT) do end
	end_visit_octal_constant (node: AST_OCTAL_CONSTANT) do end
	end_visit_parameter_list (node: AST_PARAMETER_LIST) do end
	end_visit_parenthesized_expression (node: AST_PARENTHESIZED_EXPRESSION) do end
	end_visit_pointer_type (node: AST_POINTER_TYPE) do end
	end_visit_postfix_expression (node: AST_POSTFIX_EXPRESSION) do end
	end_visit_sizeof_expression (node: AST_SIZEOF_EXPRESSION) do end
	end_visit_string_constant (node: AST_STRING_CONSTANT) do end
	end_visit_struct_declaration (node: AST_STRUCT_DECLARATION) do end
	end_visit_struct_type (node: AST_STRUCT_TYPE) do end
	end_visit_switch_case (node: AST_SWITCH_CASE) do end
	end_visit_switch_default (node: AST_SWITCH_DEFAULT) do end
	end_visit_switch_statement (node: AST_SWITCH_STATEMENT) do end
	end_visit_translation_unit (node: AST_TRANSLATION_UNIT) do end
	end_visit_typedef_declaration (node: AST_TYPEDEF_DECLARATION) do end
	end_visit_typedef_type (node: AST_TYPEDEF_TYPE) do end
	end_visit_unary_expression (node: AST_UNARY_EXPRESSION) do end
	end_visit_variable_declaration (node: AST_VARIABLE_DECLARATION) do end
	end_visit_variable_expression (node: AST_VARIABLE_EXPRESSION) do end
	end_visit_while_statement (node: AST_WHILE_STATEMENT) do end

end
