class
	OO_FINISH

inherit
	TRANSFORMATION_STEP
	NAME_UTILITIES

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			add_system_superclasses
			if not transformer.project_settings.disable_oo_reengineering then
				remove_feature_prefixes
				if not transformer.project_settings.disable_oo_structs then
					update_class_fields_where_necessary
				end
			end
			Result := True
		end

	description: STRING
		do
			Result := "Finishing up O-O reengineering"
		end

feature {NONE}

	add_system_superclasses
		local
			l_cur_data: TUPLE [struct: AST_STRUCT_DECLARATION; is_static: BOOLEAN]
			l_visitor: AST_AGENT_VISITOR
		do
			create l_cur_data
			create l_visitor
			l_visitor.agents.end_visit_function_call := agent (ag_node: AST_FUNCTION_CALL; ag_cur_data: TUPLE [struct: AST_STRUCT_DECLARATION; is_static: BOOLEAN])
				do
					if
						attached ag_node.declaration_binding as called_function and then
						called_function.oo_class.source_info.is_system_source
					then
						if ag_cur_data.is_static then
							ag_cur_data.struct.oo_static_superclasses.extend (called_function.oo_class)
						elseif not ag_cur_data.struct.oo_static_superclasses.has (called_function.oo_class) then
							ag_cur_data.struct.oo_superclasses.extend (called_function.oo_class)
						end
					end
				end (?, l_cur_data)
			l_visitor.agents.end_visit_variable_expression := agent (ag_node: AST_VARIABLE_EXPRESSION; ag_cur_data: TUPLE [struct: AST_STRUCT_DECLARATION; is_static: BOOLEAN])
				do
					if
						attached ag_node.declaration_binding as called_variable and then
						attached {AST_TRANSLATION_UNIT} called_variable.parent and then
						called_variable.oo_class.source_info.is_system_source
					then
						if ag_cur_data.is_static then
							ag_cur_data.struct.oo_static_superclasses.extend (called_variable.oo_class)
						elseif not ag_cur_data.struct.oo_static_superclasses.has (called_variable.oo_class) then
							ag_cur_data.struct.oo_superclasses.extend (called_variable.oo_class)
						end
					end
				end (?, l_cur_data)

			across transformer.tree.declarations as c loop
				if
					attached {AST_STRUCT_DECLARATION} c.item as sd and then
					sd.source_info.is_project_source
				then
					l_cur_data.struct := sd
					l_cur_data.is_static := True
					across sd.oo_static_features as c2 loop
						c2.item.accept (l_visitor)
					end
					l_cur_data.is_static := False
					across sd.oo_features as c2 loop
						c2.item.accept (l_visitor)
					end
				end
			end
		end

feature {NONE}

	remove_feature_prefixes
		do
			across transformer.tree.declarations as dc loop
				if
					attached {AST_STRUCT_DECLARATION} dc.item as sd and then
					not sd.oo_features.is_empty
				then
					remove_feature_prefixes_in_class (sd)
				end
			end
		end

	remove_feature_prefixes_in_class (a_node: AST_STRUCT_DECLARATION)
		require
			a_node /= Void
			not a_node.oo_features.is_empty
		local
			l_prefix, l_identifier: STRING_8
		do
			-- PD_MONGO_CURSOR -> mongo_cursor_
			l_prefix := a_node.identifier.substring (4, a_node.identifier.count).as_lower
			l_prefix := remove_cil_anonymous_decoration (l_prefix)
			l_prefix := pretty_print_underscores (l_prefix) + "_"

			across a_node.oo_features as c loop
				l_identifier := c.item.identifier.as_lower
				l_identifier := pretty_print_underscores (l_identifier)
				if
					l_identifier.starts_with (l_prefix) and
					l_identifier.count > l_prefix.count
				then
					--Io.put_string ("{" + a_node.identifier + "}." + c.item.identifier + " -> ")
					l_identifier.remove_head (l_prefix.count)
					c.item.identifier.make_from_string (l_identifier)
					--Io.put_string ("{" + a_node.identifier + "}." + c.item.identifier + "%N")
				end
			end
		end

feature {NONE}

	update_class_fields_where_necessary
			-- Functions that changed a struct pointer argument to Current might become
			-- inconsistent after function calls or assignments if Current is aliased
			-- somewhere else (see oo_reeng_struct_pointer_arg_aliasing.c).
			-- We therefore add calls to `update_class_fields' after assignments and function calls.
		local
			l_update_statements: HASH_SET [AST_STATEMENT]
			l_visitor: AST_AGENT_VISITOR
		do
			-- collect all the statements to avoid making duplicate calls to `update_class_fields' (e.g. from `tmp = foo()')
			create l_update_statements.make (16)
			create l_visitor
			l_visitor.agents.end_visit_function_call := agent (ag_node: AST_FUNCTION_CALL; ag_update_statements: HASH_SET [AST_STATEMENT])
				do
					-- TODO; only add call if necessary
					ag_update_statements.extend ({AST_STATEMENT} #? ag_node.parent_of_type ({AST_STATEMENT}))
				end (?, l_update_statements)
			l_visitor.agents.end_visit_assignment_expression := agent (ag_node: AST_ASSIGNMENT_EXPRESSION; ag_update_statements: HASH_SET [AST_STATEMENT])
				do
					-- TODO; can more calls be removed?
					if
						not attached {AST_VARIABLE_EXPRESSION} ag_node.target and
						not (attached {AST_EIFFEL_EXPRESSION} ag_node.target as ee and then ee.mixed.count = 1 and then ee.is_string (1, "Result"))
					then
						ag_update_statements.extend ({AST_STATEMENT} #? ag_node.parent_of_type ({AST_STATEMENT}))
					end
				end (?, l_update_statements)

			across transformer.tree.declarations as c loop
				if
					attached {AST_FUNCTION_DECLARATION} c.item as fd and then
					fd.source_info.is_project_source and then
					fd.oo_class_assignment_data.applied_heuristic = {OO_FUNCTIONS_TRANSFORMER}.Heuristic_struct_pointer_arg
				then
					fd.accept (l_visitor)
				end
			end

			across l_update_statements as c loop
				insert_adjacent_statement (
						c.item,
						create {AST_EXPRESSION_STATEMENT}.make (create {AST_EIFFEL_EXPRESSION}.make_mixed (
								<<"update_class_fields", " -- Current might have been changed; synchronize">>,
								create {AST_BASIC_TYPE}.make (Ast_bt_void)
							)),
						True
					)
			end
		end

end
