class
	PARAMETER_ASSIGNMENT_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			visit_function_declaration,
			end_visit_function_declaration,
			end_visit_variable_expression
		end
	ROUTINE_PARAMETER_UTILITIES

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description: STRING
		do
			Result := "Transforming assignments to function parameters"
		end

feature {NONE}

	variable_assignments: HASH_SET [AST_VARIABLE_DECLARATION]

feature {ANY} -- AST_DEFAULT_VISITOR

	visit_function_declaration (node: AST_FUNCTION_DECLARATION): BOOLEAN
		do
			if node.is_regular then
				create variable_assignments.make (16)
				Result := True
			end
		end

	end_visit_function_declaration (node: AST_FUNCTION_DECLARATION)
		local
			params: LIST [AST_NODE]
		do
			if node.is_regular then
				from
					params := node.parameter_list.parameters
					params.finish
				until
					params.before
				loop
					if
						attached {AST_VARIABLE_DECLARATION} params.item as orig_param and then
						variable_assignments.has (orig_param)
					then
						change_to_local (orig_param)
					end
					params.back
				end
				variable_assignments := Void
			end
		end

	end_visit_variable_expression (node: AST_VARIABLE_EXPRESSION)
		do
			if
				variable_assignments /= Void and then
				attached {AST_VARIABLE_DECLARATION} node.declaration_binding as vd and then
				(
					(attached {AST_ASSIGNMENT_EXPRESSION} node.parent as ae and then ae.target = node) or
					(attached {AST_BASIC_TYPE} node.type_binding as bt and then bt.type = Ast_bt_builtin_va_list)
				)
			then
				variable_assignments.extend (vd)
			end
		end

end
