class
	SOURCE_LOADER

inherit
	TRANSFORMATION_STEP

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		local
			source_file_name: STRING
			index: INTEGER
		do
			source_file_name := transformer.project_settings.cil_file.name
			index := source_file_name.last_index_of ('\', source_file_name.count).max (
					source_file_name.last_index_of ('/', source_file_name.count))
			source_file_name := source_file_name.substring (index + 1, source_file_name.count)
			transformer.tree.source_file_name := source_file_name

			transformer.tree.declarations.wipe_out
			setup_antlr
			Result := parse_file (create {PLAIN_TEXT_FILE}.make (transformer.project_settings.cil_file.name))
			dispose_antlr
		end

	description: STRING
		do
			Result := "Loading Source File"
		end

feature {NONE}

	parse_file (file: PLAIN_TEXT_FILE): BOOLEAN
		require
			file /= Void and then file.exists
		local
			source, tmp_source: STRING
			declaration_start, cur_newline_index, bak: INTEGER
		do
			file.open_read
			file.read_stream (file.count)
			source := file.last_string
			file.close

			from
				Result := True
				declaration_start := 1
				cur_newline_index := source.index_of ('%N', 1)
			until
				not Result or cur_newline_index = 0 or cur_newline_index = source.count
			loop
				if has_substring_at (source, "#line ", cur_newline_index + 1) then
					bak := cur_newline_index
					cur_newline_index := source.index_of ('%N', cur_newline_index + 1)
					if source[cur_newline_index + 1] /= ' ' or has_substring_at (source, " __attribute", cur_newline_index + 1) then
						Result := parse_external_declaration (source, declaration_start, bak)
						declaration_start := bak + 1
					end
				else
					if has_substring_at (source, "/* compiler builtin:", cur_newline_index + 1) then
						tmp_source := "#line 1 %""
						if {PLATFORM}.is_windows then
							tmp_source.append ("c:/FOO")
						end
						tmp_source.append ("/usr/include/gcc_builtin.h%"%N")
						tmp_source.append ("extern ")
						tmp_source.append (source.substring (cur_newline_index + 21, source.index_of (';', cur_newline_index + 21)))
						Result := parse_external_declaration (tmp_source, 1, tmp_source.count)
					end
					cur_newline_index := source.index_of ('%N', cur_newline_index + 1)
				end
			end
			if Result then
				Result := parse_external_declaration (source, declaration_start, source.count)
			end
		end

	has_substring_at (s1, s2: STRING; index: INTEGER): BOOLEAN
		require
			s1 /= Void and then not s1.is_empty
			s2 /= Void and then not s2.is_empty
			index >= 1 and index <= s1.count
		local
			s1_i, s2_i: INTEGER
		do
			if index + s2.count - 1 <= s1.count then
				Result := True
				from
					s1_i := index
					s2_i := 1
				until
					not Result or else s2_i > s2.count
				loop
					Result := s1[s1_i] = s2[s2_i]
					s1_i := s1_i + 1
					s2_i := s2_i + 1
				end
			end
		end

	parse_external_declaration (source: STRING; start_index, end_index: INTEGER): BOOLEAN
		require
			source /= Void and then not source.is_empty
			start_index >= 1 and start_index <= source.count
			end_index >= 1 and end_index <= source.count
			start_index < end_index
		local
			antlr_data, antlr_c_node: POINTER
			antlr_node: ANTLR_AST_NODE
		do
			antlr_data := antlr_parse (source.area.item_address (start_index - 1), end_index - start_index + 1)
			if antlr_data /= default_pointer then
				antlr_c_node := antlr_data_root_node (antlr_data)
				if antlr_c_node /= default_pointer then
					create antlr_node.make (antlr_c_node)
					if antlr_node.children.count > 1 and antlr_node.children.count <= 10 then
						log ("Warning: Found " + antlr_node.children.count.out + " external declarations (expected only one):")
						log (source.substring (start_index, end_index))
					elseif antlr_node.children.count > 10 then
						log ("Warning: Found " + antlr_node.children.count.out + " external declarations (expected only one)")
					end
					Result := extend_tree (antlr_node)
				else
					log ("Syntax error in:")
					log (source.substring (start_index, end_index))
					log (antlr_get_error)
				end
				antlr_data_dispose (antlr_data)
			else
				log ("ANTLR setup failed! Out of memory?")
			end
		end

	extend_tree (a_antlr_node: ANTLR_AST_NODE): BOOLEAN
		require
			a_antlr_node /= Void and then a_antlr_node.type = Ast_translation_unit
		local
			l_source_info: SOURCE_INFO
		do
			Result := True
			across a_antlr_node.children as c loop
				create l_source_info.make (antlr_source_file (c.item.antlr_c_node))
				inspect
					c.item.type
				when C_kw_typedef then
					transformer.tree.declarations.extend (create {AST_TYPEDEF_DECLARATION}.make_from_antlr (c.item, transformer.tree, l_source_info))
				when C_kw_struct, C_kw_union then
					transformer.tree.declarations.extend (create {AST_STRUCT_DECLARATION}.make_from_antlr (c.item, transformer.tree, l_source_info))
				when C_kw_enum then
					transformer.tree.declarations.extend (create {AST_ENUM_DECLARATION}.make_from_antlr (c.item, transformer.tree, l_source_info))
				when Ast_variable_declaration_or_type then
					transformer.tree.declarations.extend (create {AST_VARIABLE_DECLARATION}.make_from_antlr (c.item, transformer.tree, l_source_info))
				when Ast_function_declaration then
					transformer.tree.declarations.extend (create {AST_FUNCTION_DECLARATION}.make_from_antlr (c.item, transformer.tree, l_source_info))
				end
			end
		end

feature {NONE} -- ANTLR

	setup_antlr
		external
			"C inline use <CLexer.h>, <CParser.h>"
		alias
			"setup_typedef_set();"
		end

	dispose_antlr
		external
			"C inline use <CLexer.h>, <CParser.h>"
		alias
			"dispose_typedef_set();"
		end

	antlr_parse (str: POINTER; strlen: INTEGER): POINTER
		external
			"C inline use <CLexer.h>, <CParser.h>"
		alias
			"[
				struct antlr_data_struct
				{
					pANTLR3_INPUT_STREAM input;
					pCLexer lxr;
					pANTLR3_COMMON_TOKEN_STREAM tstream;
					pCParser psr;
					pANTLR3_BASE_TREE root_node;
				};
				struct antlr_data_struct* antlr_data = (struct antlr_data_struct*)malloc(sizeof(struct antlr_data_struct));
				CParser_translation_unit_return langAST;

				antlr_data->input = antlr3NewAsciiStringInPlaceStream((pANTLR3_UINT8)$str, $strlen, (pANTLR3_UINT8)"C2Eif Input");
				if(antlr_data->input == NULL)
				{
					ANTLR3_FPRINTF(stderr, "ANTLR: antlr3NewAsciiStringInPlaceStream() failed\n");
					return NULL;
				}

				antlr_data->lxr = CLexerNew(antlr_data->input);
				if(antlr_data->lxr == NULL)
				{
					ANTLR3_FPRINTF(stderr, "ANTLR: CLexerNew() failed\n");
					return NULL;
				}

				antlr_data->tstream = antlr3CommonTokenStreamSourceNew(ANTLR3_SIZE_HINT, antlr_data->lxr->pLexer->rec->state->tokSource);
				if(antlr_data->tstream == NULL)
				{
					ANTLR3_FPRINTF(stderr, "ANTLR: antlr3CommonTokenStreamSourceNew() failed\n");
					return NULL;
				}

				antlr_data->psr = CParserNew(antlr_data->tstream);
				if(antlr_data->psr == NULL)
				{
					ANTLR3_FPRINTF(stderr, "ANTLR: CParserNew() failed\n");
					return NULL;
				}

				langAST = antlr_data->psr->translation_unit(antlr_data->psr);
				antlr_data->root_node = (antlr_data->psr->pParser->rec->getNumberOfSyntaxErrors(antlr_data->psr->pParser->rec) == 0) ? langAST.tree : NULL;
				return antlr_data;
			]"
		end

	antlr_data_root_node (a_antlr_data: POINTER): POINTER
		require
			a_antlr_data /= default_pointer
		external
			"C inline use <CLexer.h>, <CParser.h>"
		alias
			"[
				struct antlr_data_struct
				{
					pANTLR3_INPUT_STREAM input;
					pCLexer lxr;
					pANTLR3_COMMON_TOKEN_STREAM tstream;
					pCParser psr;
					pANTLR3_BASE_TREE root_node;
				};
				struct antlr_data_struct* antlr_data = (struct antlr_data_struct*)$a_antlr_data;
				return antlr_data->root_node;
			]"
		end

	antlr_get_error: STRING
		external
			"C inline use <error_handlers.h>"
		alias
			"return eh_dispose_error_string();"
		end

	antlr_data_dispose (a_antlr_data: POINTER)
		require
			a_antlr_data /= default_pointer
		external
			"C inline use <CLexer.h>, <CParser.h>"
		alias
			"[
				struct antlr_data_struct
				{
					pANTLR3_INPUT_STREAM input;
					pCLexer lxr;
					pANTLR3_COMMON_TOKEN_STREAM tstream;
					pCParser psr;
					pANTLR3_BASE_TREE root_node;
				};
				struct antlr_data_struct* antlr_data = (struct antlr_data_struct*)$a_antlr_data;
				antlr_data->psr->free(antlr_data->psr);
				antlr_data->tstream->free(antlr_data->tstream);
				antlr_data->lxr->free(antlr_data->lxr);
				antlr_data->input->close(antlr_data->input);
				free(antlr_data);
			]"
		end

	antlr_source_file (a_antlr_c_node: POINTER): STRING
		require
			a_antlr_c_node /= default_pointer
		external
			"C inline use <CLexer.h>, <CParser.h>"
		alias
			"[
				pANTLR3_BASE_TREE node = $a_antlr_c_node;
				EIF_REFERENCE ret;
				char *s;
				
				if(node->u == NULL) return NULL;
				
				s = node->u;
				while(*s != '"') s++;
				*s = '\0';
				ret = eif_string(node->u);
				*s = '"';
				return ret;
			]"
		end

end
