class
	STRUCT_WRITER_C

inherit
	STRUCT_WRITER_OO
		redefine
			write_class,
			write_make,
			write_inherit_clause,
			write_struct_fields,
			write_struct_field_setters,
			write_copy_is_equal,
			write_structure_info,
			write_bitfield_helper
		end

create
	make

feature {NONE}

	c_include: STRING_8
		-- e.g `"xeyes.h"'

feature {NONE}

	write_class
		do
			if
				not node.source_info.is_pure_oo or
				not node.oo_features.is_empty
			then
				c_include := "%%%"" + c_header_file ({AST_TRANSLATION_UNIT} #? node.parent) + "%%%""

				Precursor
				buffer.remove_tail (4) -- remove "end%N"

				write_c_field_getters
				write_c_field_setters
				write_offset_of_query

				buffer.append ("end%N")
			end
		end

	write_inherit_clause
		local
			l_first: BOOLEAN
		do
			if not node.source_info.is_pure_oo then
				buffer.append ("inherit%N")
				if
					not node.oo_static_features.is_empty or
					across node.oo_features as c some attached {AST_FUNCTION_DECLARATION} c.item as fd and then fd.needs_address end
				then
					buffer.append ("	" + node.identifier + "_STATIC%N")
				else
					buffer.append ("	CE_UTILITIES%N")
				end
				if
					node.oo_superclasses.is_empty or
					across node.oo_superclasses as c all c.item.source_info.is_pure_oo end
				then
					buffer.append ("		redefine%N")
				else
					buffer.append ("		undefine%N")
				end
				buffer.append ("			copy,%N")
				buffer.append ("			is_equal%N")
				buffer.append ("		end%N")
				if
					node.oo_superclasses.is_empty or
					across node.oo_superclasses as c all c.item.source_info.is_pure_oo end
				then
					buffer.append ("	CE_CLASS%N")
					buffer.append ("		redefine%N")
					buffer.append ("			make,%N")
					buffer.append ("			make_shared,%N")
					buffer.append ("			make_initialized,%N")
					buffer.append ("			structure_size,%N")
					buffer.append ("			offset_of,%N")
					buffer.append ("			copy,%N")
					buffer.append ("			is_equal%N")
					buffer.append ("		end%N")
					across node.oo_superclasses as c loop
						buffer.append ("	" + c.item.identifier + "_STATIC%N")
						buffer.append ("		redefine%N")
						buffer.append ("			copy,%N")
						buffer.append ("			is_equal%N")
						buffer.append ("		end%N")
					end
				else
					across node.oo_superclasses as c loop
						buffer.append ("	" + c.item.identifier)
						if c.item.source_info.is_pure_oo then
							buffer.append ("_STATIC%N")
							buffer.append ("		undefine%N")
							buffer.append ("			copy,%N")
							buffer.append ("			is_equal%N")
							buffer.append ("		end%N")
						else
							buffer.append ("%N")

							-- rename
							l_first := True
							node.fields.start
							across c.item.fields as c2 loop
								if
									attached c2.item.identifier as super_identifier and then
									attached node.fields.item.identifier as sub_identifier and then
									not super_identifier.is_equal (sub_identifier)
								then
									if l_first then
										buffer.append ("		rename%N")
										l_first := False
									else
										buffer.append (",%N")
									end
									buffer.append ("			" + super_identifier + " as " + sub_identifier)
								end
								node.fields.forth
							end
							if not l_first then
								buffer.append ("%N")
							end

							-- redefine
							buffer.append ("		redefine%N")
							buffer.append ("			structure_size,%N")
							buffer.append ("			structure_alignment,%N")
							buffer.append ("			c__offset_of%N")
							buffer.append ("		end%N")
						end
					end
				end
				if regular_class_needs_asm then
					buffer.append ("	" + eiffel_class_name (({AST_TRANSLATION_UNIT} #? node.parent).source_file_name) + "_ASSEMBLER%N")
					buffer.append ("		redefine%N")
					buffer.append ("			copy,%N")
					buffer.append ("			is_equal%N")
					buffer.append ("		end%N")
				end
				buffer.append ("%N")
			end
		end

	write_make
		do
			if not node.source_info.is_pure_oo then
				if
					node.oo_superclasses.is_empty or
					across node.oo_superclasses as c all c.item.source_info.is_pure_oo end
				then
					buffer.append ("feature {NONE}%N%N")

					buffer.append ("	make%N")
					buffer.append ("		do%N")
					buffer.append ("			internal_c_pointer := internal_c_pointer.memory_alloc (structure_size)%N")
					buffer.append ("			is_allocated_memory := True%N")
					buffer.append ("		end%N%N")

					buffer.append ("	make_shared (a_c_pointer: POINTER)%N")
					buffer.append ("		do%N")
					buffer.append ("			internal_c_pointer := a_c_pointer%N")
					buffer.append ("		end%N%N")

					buffer.append ("	make_initialized (a_data: TUPLE [%N")
					across node.fields as c loop
						if attached c.item.identifier as identifier then
							buffer.append ("				a_" + identifier + ": " + c.item.type.eiffel_code + "%N")
						end
					end
					buffer.append ("			])%N")
					buffer.append ("		require else%N")
					buffer.append ("			a_data /= Void%N")
					buffer.append ("		do%N")
					buffer.append ("			make%N")
					across node.fields as c loop
						if attached c.item.identifier as identifier then
							if attached {AST_ARRAY_TYPE} c.item.type then
								buffer.append ("			" + identifier + ".c_pointer.memory_copy (a_data.a_" + identifier + ".c_pointer, ")
								buffer.append ("a_data.a_" + identifier + ".total_count * a_data.a_" + identifier + ".element_size)%N")
							elseif attached {AST_STRUCT_TYPE} c.item.type then
								buffer.append ("			" + identifier + ".c_pointer.memory_copy (a_data.a_" + identifier + ".c_pointer, ")
								buffer.append ("a_data.a_" + identifier + ".structure_size)%N")
							else
								buffer.append ("			Current." + identifier + " := a_data.a_" + identifier + "%N")
							end
						end
					end
					buffer.append ("		end%N%N")
				end
			end
		end

	write_struct_fields
		local
			l_first: BOOLEAN
			l_cur_at: AST_ARRAY_TYPE
		do
			if not node.source_info.is_pure_oo then
				l_first := True
				across fields as c loop
					if attached c.item.identifier as identifier then
						if l_first then
								buffer.append ("feature {ANY} -- fields%N%N")
							l_first := False
						end

						buffer.append ("	" + identifier + ": ")
						buffer.append (c.item.type.eiffel_code)
						if
							not attached {AST_ARRAY_TYPE} c.item.type and
							not (
									attached {AST_STRUCT_TYPE} c.item.type as st and then
									attached {AST_STRUCT_DECLARATION} st.declaration_binding as sd and then
									sd.typedef_name = Void and then
									st.declaration_binding.orig_c_identifier.starts_with ("__anon")
								)
						then
							buffer.append (" assign set__" + identifier)
						end
						buffer.append ("%N")
						if attached {AST_ARRAY_TYPE} c.item.type or attached {AST_STRUCT_TYPE} c.item.type then
							buffer.append ("		once (%"OBJECT%")%N")
						else

							buffer.append ("		do%N")
						end
						if attached {AST_POINTER_TYPE} c.item.type as pt then
							buffer.append ("			create ")
							if not attached {AST_FUNCTION_TYPE} pt.type then
								buffer.append ("{")
								buffer.append (effective_type (pt))
								buffer.append ("} ")
							end
							buffer.append ("Result.make_shared (c__" + identifier + " (internal_c_pointer))%N")
						elseif attached {AST_ARRAY_TYPE} c.item.type as at then
							buffer.append ("			create {" + effective_type (at) + "} Result.make_shared (c__" + identifier + " (internal_c_pointer), <<")
							from
								l_cur_at := at
							until
								l_cur_at = Void
							loop
								if l_cur_at /= at then
									buffer.append (",")
								end
								if l_cur_at.dimension /= Void then
									buffer.append (l_cur_at.dimension.eiffel_code)
								else
									-- flexible array member at end of struct; no dimension needed
									buffer.append ("0")
								end
								l_cur_at ?= l_cur_at.type
							end
							buffer.append (">>)%N")
						elseif
							attached {AST_STRUCT_TYPE} c.item.type or
							(attached {AST_BASIC_TYPE} c.item.type as bt and then bt.type = ast_bt_builtin_va_list)
						then
							buffer.append ("			create Result.make_shared (c__" + identifier + " (internal_c_pointer))%N")
						elseif attached {AST_BASIC_TYPE} c.item.type as bt and then bt.type = Ast_bt_long_double then
							buffer.append ("			Result.c_pointer.memory_copy (c__" + identifier + " (internal_c_pointer), {CE_UTILITIES}.Real_96_bytes)%N")
						else
							buffer.append ("			Result := c__" + identifier + " (internal_c_pointer)%N")
						end
						buffer.append ("		end%N%N")
					end
				end
			end
		end

	write_struct_field_setters
		local
			l_first: BOOLEAN
		do
			if not node.source_info.is_pure_oo then
				l_first := True
				across fields as c loop
					if
						attached c.item.identifier as identifier and
						not attached {AST_ARRAY_TYPE} c.item.type and
						not (
								attached {AST_STRUCT_TYPE} c.item.type as st and then
								attached {AST_STRUCT_DECLARATION} st.declaration_binding as sd and then
								sd.typedef_name = Void and then
								st.declaration_binding.orig_c_identifier.starts_with ("__anon")
							)
					then
						if l_first then
							buffer.append ("feature {ANY} -- field setters%N%N")
							l_first := False
						end

						buffer.append ("	set__" + identifier + " (a_" + identifier + ": ")
						buffer.append (c.item.type.eiffel_code)
						buffer.append (")%N")
						buffer.append ("		do%N")
						buffer.append ("			c__set_" + identifier + " (internal_c_pointer, a_" + identifier)
						if
							attached {AST_POINTER_TYPE} c.item.type or
							attached {AST_STRUCT_TYPE} c.item.type or
							(attached {AST_BASIC_TYPE} c.item.type as bt and then bt.type = Ast_bt_long_double) or
							(attached {AST_BASIC_TYPE} c.item.type as bt and then bt.type = ast_bt_builtin_va_list)
						then
							buffer.append (".c_pointer")
						end
						buffer.append (")%N")
						buffer.append ("		end%N%N")
					end
				end
			end
		end

	write_copy_is_equal
		do
			if not node.source_info.is_pure_oo then
				buffer.append ("feature {ANY}%N%N")

				buffer.append ("	structure_size: INTEGER_32%N")
				buffer.append ("		external%N")
				buffer.append ("			%"C inline use " + c_include + "%"%N")
				buffer.append ("		alias%N")
				buffer.append ("			%"return sizeof(" + c_name (node) + ");%"%N")
				buffer.append ("		end%N%N")

				buffer.append ("	structure_alignment: INTEGER_32%N")
				buffer.append ("		external%N")
				buffer.append ("			%"C inline use " + c_include + "%"%N")
				buffer.append ("		alias%N")
				buffer.append ("			%"return __alignof__(" + c_name (node) + ");%"%N")
				buffer.append ("		end%N%N")

				if
					node.oo_superclasses.is_empty or
					across node.oo_superclasses as c all c.item.source_info.is_pure_oo end
				then
					buffer.append ("	offset_of (a_field_name: STRING_8): INTEGER_32%N")
					buffer.append ("		local%N")
					buffer.append ("			l_c_str: C_STRING%N")
					buffer.append ("		do%N")
					buffer.append ("			create l_c_str.make (a_field_name)%N")
					buffer.append ("			Result := c__offset_of (l_c_str.item)%N")
					buffer.append ("		end%N%N")

					buffer.append ("feature {ANY}%N%N")

					buffer.append ("	copy (a_other: like Current)%N")
					buffer.append ("		do%N")
					buffer.append ("			make%N")
					buffer.append ("			internal_c_pointer.memory_copy (a_other.c_pointer, structure_size)%N")
					buffer.append ("		end%N%N")

					buffer.append ("	is_equal (a_other: like Current): BOOLEAN%N")
					buffer.append ("		do%N")
					buffer.append ("			Result := internal_c_pointer.memory_compare (a_other.c_pointer, structure_size)%N")
					buffer.append ("		end%N%N")
				end
			end
		end

	write_structure_info
		do
			if not node.source_info.is_pure_oo then
				buffer.append ("feature {NONE}%N%N")

				if
					node.oo_superclasses.is_empty or
					across node.oo_superclasses as c all c.item.source_info.is_pure_oo end
				then
					buffer.append ("	structure_info: CE_STRUCTURE%N")
					buffer.append ("		do%N")
					buffer.append ("			-- dummy for CE_CLASS invariant%N")
					if node.is_union then
						buffer.append ("			create Result.make_union (<<>>)%N")
					else
						buffer.append ("			create Result.make_struct (<<>>)%N")
					end
						buffer.append ("			Result.alignment := structure_alignment%N")
					buffer.append ("		end%N%N")
				end
			end
		end

	write_bitfield_helper
		do
			-- not needed
		end

	write_c_field_getters
		do
			across fields as c loop
				if attached c.item.identifier as identifier then
					buffer.append ("	c__" + identifier + " (a_c_pointer: POINTER): " + external_type (c.item.type) + "%N")
					buffer.append ("		external%N")
					buffer.append ("			%"C inline use " + c_include + "%"%N")
					buffer.append ("		alias%N")
					buffer.append ("			%"return ")
					if
						attached {AST_POINTER_TYPE} c.item.type or
						attached {AST_ARRAY_TYPE} c.item.type or
						attached {AST_STRUCT_TYPE} c.item.type or
						(attached {AST_BASIC_TYPE} c.item.type as bt and then bt.type = Ast_bt_long_double)
					then
						buffer.append ("(EIF_POINTER)")
					end
					if
						attached {AST_STRUCT_TYPE} c.item.type or
						(attached {AST_BASIC_TYPE} c.item.type as bt and then bt.type = Ast_bt_long_double)
					then
						buffer.append ("&")
					end
					buffer.append ("((" + c_name (node) + " *)$a_c_pointer)->" + c.item.orig_c_identifier + ";%"%N")
					buffer.append ("		end%N%N")
				end
			end
		end

	write_c_field_setters
		do
			across fields as c loop
				if
					attached c.item.identifier as identifier and
					not attached {AST_ARRAY_TYPE} c.item.type and
					not (
							attached {AST_STRUCT_TYPE} c.item.type as st and then
							attached {AST_STRUCT_DECLARATION} st.declaration_binding as sd and then
							sd.typedef_name = Void and then
							st.declaration_binding.orig_c_identifier.starts_with ("__anon")
						)
				then
					buffer.append ("	c__set_" + identifier + " (a_c_pointer: POINTER; a_" + identifier + ": " + external_type (c.item.type) + ")%N")
					buffer.append ("		external%N")
					buffer.append ("			%"C inline use " + c_include + "%"%N")
					buffer.append ("		alias%N")
					buffer.append ("			%"((" + c_name (node) + " *)$a_c_pointer)->" + c.item.orig_c_identifier + " = ")
					if
						attached {AST_STRUCT_TYPE} c.item.type as st and then
						attached st.declaration_binding as sd
					then
						buffer.append ("*(" + c_name (sd) + " *)")
					elseif attached {AST_BASIC_TYPE} c.item.type as bt and then bt.type = Ast_bt_long_double then
						buffer.append ("*(long double *)")
					end
					buffer.append ("$a_" + identifier.as_lower + ";%"%N")
					buffer.append ("		end%N%N")
				end
			end
		end

	write_offset_of_query
		do
			buffer.append ("	c__offset_of (a_field_name: POINTER): INTEGER_32%N")
			buffer.append ("		external%N")
			buffer.append ("			%"C inline use " + c_include + ", <string.h>, <stddef.h>, <stdio.h>%"%N")
			buffer.append ("		alias%N")
			buffer.append ("			%"[%N")
			across node.fields as c loop
				if
					attached c.item.identifier as identifier and
					not attached {AST_BITFIELD} c.item
				then
					buffer.append ("				if(strcmp($a_field_name, %"" + identifier + "%") == 0) return offsetof(" + c_name (node) + ", " + c.item.orig_c_identifier + ");%N")
				end
			end
			buffer.append ("				fprintf(stderr, %"" + c_name (node) + ".%%s not found\n%", $a_field_name);%N")
			buffer.append ("				return -1;%N")
			buffer.append ("			]%"%N")
			buffer.append ("		end%N%N")
		end

feature {NONE}

	c_name (a_node: AST_STRUCT_DECLARATION): STRING_8
		do
			if
				a_node.source_info.is_system_source and
				a_node.orig_c_identifier.starts_with ("__anon") and
				a_node.typedef_name /= Void
			then
				Result := a_node.typedef_name
			else
				if a_node.is_union then
					Result := "union " + a_node.orig_c_identifier
				else
					Result := "struct " + a_node.orig_c_identifier
				end
			end
		end

	external_type (a_type: AST_TYPE): STRING_8
		do
			if
				attached {AST_POINTER_TYPE} a_type or
				attached {AST_ARRAY_TYPE} a_type or
				attached {AST_STRUCT_TYPE} a_type or
				(attached {AST_BASIC_TYPE} a_type as bt and then bt.type = Ast_bt_long_double) or
				(attached {AST_BASIC_TYPE} a_type as bt and then bt.type = ast_bt_builtin_va_list)
			then
				Result := "POINTER"
			else
				Result := a_type.eiffel_code
			end
		end

end
