class
	AST_C_CREATOR

inherit
	AST_VISITOR
	AST_UTILITIES

create
	make

feature {NONE}

	make
		do
			create buffer.make_empty
		end

feature {ANY}

	buffer: STRING

feature {ANY}

	visit_translation_unit (node: AST_TRANSLATION_UNIT): BOOLEAN
		do
			across node.declarations as c loop
				if not c.item.source_info.is_pure_oo then
					buffer.append ("#line %"" + c.item.source_info.orig_source_file + "%"%N")
				end
				c.item.accept (Current)
				if attached {AST_VARIABLE_DECLARATION} c.item then
					buffer.append (";")
				end
				buffer.append ("%N%N")
			end
		end

	visit_typedef_declaration (node: AST_TYPEDEF_DECLARATION): BOOLEAN
		do
			buffer.append ("typedef ")
			if attached {AST_FUNCTION_TYPE} node.type as tmp then
				(create {AST_FUNCTION_DECLARATION}.make_forward_declaration (tmp.return_type.ast_twin, node.identifier, tmp.parameter_list.ast_twin)).accept (Current)
			else
				(create {AST_VARIABLE_DECLARATION}.make (node.type.ast_twin, node.identifier, Void)).accept (Current)
				buffer.append (";")
			end
		end

	visit_struct_declaration (node: AST_STRUCT_DECLARATION): BOOLEAN
		do
			if node.is_union then
				buffer.append ("union ")
			else
				buffer.append ("struct ")
			end
			buffer.append (node.orig_c_identifier)
			if not node.is_forward_declaration then
				buffer.append (" {%N")
				indent := indent + 1
				from
					node.fields.start
				until
					node.fields.after
				loop
					print_indent
					node.fields.item.accept (Current)
					buffer.append (";%N")
					node.fields.forth
				end
				indent := indent - 1
				buffer.append ("}")
			end
			if node.gcc_attribute_specifier /= Void then
				buffer.append (" ")
				node.gcc_attribute_specifier.accept (Current);
			end
			buffer.append (";")
		end

	visit_enum_declaration (node: AST_ENUM_DECLARATION): BOOLEAN
		do
			buffer.append ("enum ")
			buffer.append (node.identifier)
			buffer.append (" {%N")
			indent := indent + 1
			across node.fields as c loop
				print_indent
				buffer.append (c.item.orig_c_identifier)
				buffer.append (" = ")
				c.item.initializer.accept (Current)
				if c.target_index < node.fields.count then
					buffer.append (",")
				end
				buffer.append ("%N")
			end
			indent := indent - 1
			buffer.append ("};")
		end

	visit_variable_declaration (node: AST_VARIABLE_DECLARATION): BOOLEAN
		do
			if not node.storage_class_specifiers.is_empty then
				print_storage_class_specifiers (node.storage_class_specifiers)
				buffer.append (" ")
			end

			if attached {AST_POINTER_TYPE} node.type as pt and then attached {AST_ARRAY_TYPE} pt.non_pointer_type then
				print_array_pointer (pt, node.orig_c_identifier)
			elseif attached {AST_ARRAY_TYPE} node.type as array_type then
				print_function_pointer_head (array_type.non_array_type)
				buffer.append (" ")
				buffer.append (node.orig_c_identifier)
				array_type.accept (Current)
				print_function_pointer_tail (array_type.non_array_type)
			else
				print_function_pointer_head (node.type)
				buffer.append (" ")
				if node.gcc_attribute_specifier /= Void then
					node.gcc_attribute_specifier.accept (Current)
					buffer.append (" ")
				end
				buffer.append (node.orig_c_identifier)
				print_function_pointer_tail (node.type)
			end

			if node.initializer /= Void then
				buffer.append (" = ")
				node.initializer.accept (Current)
			end
		end

	visit_function_declaration (node: AST_FUNCTION_DECLARATION): BOOLEAN
		do
			print_storage_class_specifiers (node.storage_class_specifiers)
			if not node.storage_class_specifiers.is_empty then
				buffer.append (" ")
			end
			print_function_pointer_head (node.return_type)
			buffer.append (" ")
			buffer.append (node.orig_c_identifier)
			buffer.append ("(")
			node.parameter_list.accept (Current)
			buffer.append (")")
			print_function_pointer_tail (node.return_type)
			if node.is_regular then
				buffer.append ("%N{%N")
				indent := indent + 1
				across node.locals as declaration_cursor loop
					print_indent
					declaration_cursor.item.accept (Current)
					buffer.append (";%N")
				end
				print_indent
				node.statements.accept (Current)
				indent := indent - 1
				buffer.append ("%N}")
			else
				buffer.append (";")
			end
		end

	visit_compound_statement (node: AST_COMPOUND_STATEMENT): BOOLEAN
		do
			buffer.append ("{%N")
			indent := indent + 1
			across node.statements as c loop
				print_indent
				c.item.accept (Current)
				buffer.append ("%N")
			end
			indent := indent - 1
			print_indent
			buffer.append ("}")
		end

	visit_jump_statement (node: AST_JUMP_STATEMENT): BOOLEAN
		do
			inspect
				node.type
			when C_kw_goto then
				buffer.append ("goto ")
				buffer.append (node.goto_identifier)
			when C_kw_continue then
				buffer.append ("continue")
			when C_kw_break then
				buffer.append ("break")
			when C_kw_return then
				buffer.append ("return")
				if node.return_expression /= Void then
					buffer.append (" ")
					node.return_expression.accept (Current)
				end
			end
			buffer.append (";")
		end

	visit_expression_statement (node: AST_EXPRESSION_STATEMENT): BOOLEAN
		do
			if node.expression /= Void then
				node.expression.accept (Current)
			end
			buffer.append (";")
		end

	visit_hex_constant (node: AST_HEX_CONSTANT): BOOLEAN
		do
			buffer.append (node.value)
			inspect
				node.type_binding.type
			when Ast_bt_int then
				-- no suffix
			when Ast_bt_long then
				buffer.append ("l")
			when Ast_bt_long_long then
				buffer.append ("ll")
			when Ast_bt_unsigned_int then
				buffer.append ("u")
			when Ast_bt_unsigned_long then
				buffer.append ("ul")
			when Ast_bt_unsigned_long_long then
				buffer.append ("ull")
			end
		end

	visit_octal_constant (node: AST_OCTAL_CONSTANT): BOOLEAN
		do
			buffer.append (node.value)
			inspect
				node.type.type
			when Ast_bt_int then
				-- no suffix
			when Ast_bt_long then
				buffer.append ("l")
			when Ast_bt_long_long then
				buffer.append ("ll")
			when Ast_bt_unsigned_int then
				buffer.append ("u")
			when Ast_bt_unsigned_long then
				buffer.append ("ul")
			when Ast_bt_unsigned_long_long then
				buffer.append ("ull")
			end
		end

	visit_decimal_constant (node: AST_DECIMAL_CONSTANT): BOOLEAN
		do
			buffer.append (node.value)
			inspect
				node.type_binding.type
			when Ast_bt_int then
				-- no suffix
			when Ast_bt_long then
				buffer.append ("l")
			when Ast_bt_long_long then
				buffer.append ("ll")
			when Ast_bt_unsigned_int then
				buffer.append ("u")
			when Ast_bt_unsigned_long then
				buffer.append ("ul")
			when Ast_bt_unsigned_long_long then
				buffer.append ("ull")
			end
		end

	visit_character_constant (node: AST_CHARACTER_CONSTANT): BOOLEAN
		do
			buffer.append ("'")
			buffer.append (node.value)
			buffer.append ("'")
		end

	visit_string_constant (node: AST_STRING_CONSTANT): BOOLEAN
		do
			buffer.append ("%"")
			buffer.append (node.value)
			buffer.append ("%"")
		end

	visit_floating_point_constant (node: AST_FLOATING_POINT_CONSTANT): BOOLEAN
		do
			buffer.append (node.value)
			inspect
				node.type_binding.type
			when Ast_bt_float then
				buffer.append ("f")
			when Ast_bt_double then
				-- no suffix
			when Ast_bt_long_double then
				buffer.append ("l")
			end
		end

	visit_postfix_expression (node: AST_POSTFIX_EXPRESSION): BOOLEAN
		do
			node.main_expression.accept (Current)
			inspect
				node.type
			when C_op_open_square_bracket then
				buffer.append ("[")
				node.array_expression.accept (Current)
				buffer.append ("]")
			when C_op_point then
				buffer.append (".")
				buffer.append (node.identifier)
			when C_op_arrow then
				buffer.append ("->")
				buffer.append (node.identifier)
			when C_op_increment then
				buffer.append ("++")
			when C_op_decrement then
				buffer.append ("--")
			end
		end

	visit_parenthesized_expression (node: AST_PARENTHESIZED_EXPRESSION): BOOLEAN
		do
			buffer.append ("(")
			node.expression.accept (Current)
			buffer.append (")")
		end

	visit_variable_expression (node: AST_VARIABLE_EXPRESSION): BOOLEAN
		do
			buffer.append (node.identifier)
		end

	visit_parameter_list (node: AST_PARAMETER_LIST): BOOLEAN
		do
			from
				node.parameters.start
			until
				node.parameters.after
			loop
				node.parameters.item.accept (Current)
				if node.is_vararg or not node.parameters.islast then
					buffer.append (", ")
				end
				node.parameters.forth
			end
			if node.is_vararg then
				buffer.append ("...")
			end
		end

	visit_if_statement (node: AST_IF_STATEMENT): BOOLEAN
		do
			buffer.append ("if(")
			node.condition.accept (Current)
			buffer.append (") ")
			node.then_statements.accept (Current)
			if node.else_statements /= Void then
				buffer.append ("else")
				node.else_statements.accept (Current)
			end
		end

	visit_generic_expression (node: AST_GENERIC_EXPRESSION): BOOLEAN
		do
			node.left.accept (Current)
			inspect
				node.type
			when C_op_plus then
				buffer.append (" + ")
			when C_op_minus then
				buffer.append (" - ")
			when C_op_mult then
				buffer.append (" * ")
			when C_op_div then
				buffer.append (" / ")
			when C_op_modulo then
				buffer.append (" %% ")
			when C_op_logical_and then
				buffer.append (" && ")
			when C_op_logical_or then
				buffer.append (" || ")
			when C_op_smaller then
				buffer.append (" < ")
			when C_op_greater then
				buffer.append (" > ")
			when C_op_smaller_equal then
				buffer.append (" <= ")
			when C_op_greater_equal then
				buffer.append (" >= ")
			when C_op_equal then
				buffer.append (" == ")
			when C_op_not_equal then
				buffer.append (" != ")
			when C_op_bitand then
				buffer.append (" & ")
			when C_op_bitor then
				buffer.append (" | ")
			when C_op_comma then
				buffer.append (", ")
			when C_op_left_shift then
				buffer.append (" << ")
			when C_op_right_shift then
				buffer.append (" >> ")
			when C_op_exor then
				buffer.append (" ^ ")
			end
			node.right.accept (Current)
		end

	visit_cast_expression (node: AST_CAST_EXPRESSION): BOOLEAN
		do
			buffer.append ("(")
			node.type.accept (Current)
			buffer.append (")")
			node.expression.accept (Current)
		end

	visit_assignment_expression (node: AST_ASSIGNMENT_EXPRESSION): BOOLEAN
		do
			node.target.accept (Current)
			inspect
				node.type
			when C_op_assign then
				buffer.append (" = ")
			when C_op_mult_assign then
				buffer.append (" *= ")
			when C_op_div_assign then
				buffer.append (" /= ")
			when C_op_modulo_assign then
				buffer.append (" %%= ")
			when C_op_plus_assign then
				buffer.append (" += ")
			when C_op_minus_assign then
				buffer.append (" -= ")
			when C_op_left_shift_assign then
				buffer.append (" <<= ")
			when C_op_right_shift_assign then
				buffer.append (" >>= ")
			when C_op_bitand_assign then
				buffer.append (" &= ")
			when C_op_exor_assign then
				buffer.append (" ^= ")
			when C_op_bitor_assign then
				buffer.append (" |= ")
			end
			node.source.accept (Current)
		end

	visit_unary_expression (node: AST_UNARY_EXPRESSION): BOOLEAN
		do
			inspect
				node.type
			when C_op_increment then
				buffer.append ("++")
			when C_op_decrement then
				buffer.append ("--")
			when C_op_bitand then
				buffer.append ("&")
			when C_op_mult then
				buffer.append ("*")
			when C_op_plus then
				buffer.append ("+")
			when C_op_minus then
				buffer.append ("-")
			when C_op_neg then
				buffer.append ("~")
			when C_op_not then
				buffer.append ("!")
			end
			node.expression.accept (Current)
		end

	visit_while_statement (node: AST_WHILE_STATEMENT): BOOLEAN
		do
			buffer.append ("while(")
			node.condition.accept (Current)
			buffer.append (") ")
			node.statements.accept (Current)
		end

	visit_labeled_statement (node: AST_LABELED_STATEMENT): BOOLEAN
		do
			buffer.append (node.identifier)
			buffer.append (":%N")
			print_indent
			node.statement.accept (Current)
		end

	visit_function_call (node: AST_FUNCTION_CALL): BOOLEAN
		do
			if node.identifier_str /= Void then
				buffer.append (node.identifier_str)
			else
				buffer.append ("(")
				node.identifier_expr.accept (Current)
				buffer.append (")")
			end
			buffer.append ("(")
			from
				node.parameters.start
			until
				node.parameters.after
			loop
				if not node.parameters.isfirst then
					buffer.append (", ")
				end
				node.parameters.item.accept (Current)
				node.parameters.forth
			end
			buffer.append (")")
		end

	visit_builtin_va_arg_call (node: AST_BUILTIN_VA_ARG_CALL): BOOLEAN
		do
			buffer.append ("__builtin_va_arg(")
			node.expression.accept (Current)
			buffer.append (", ")
			node.type.accept (Current)
			buffer.append (")")
		end

	visit_alignof_call (node: AST_ALIGNOF_CALL): BOOLEAN
		do
			buffer.append ("__alignof__(")
			node.type.accept (Current)
			buffer.append (")")
		end

	visit_sizeof_expression (node: AST_SIZEOF_EXPRESSION): BOOLEAN
		do
			buffer.append ("sizeof(")
			if attached {AST_ARRAY_TYPE} node.node as at then
				at.non_array_type.accept (Current)
			end
			node.node.accept (Current)
			buffer.append (")")
		end

	visit_eiffel_expression (node: AST_EIFFEL_EXPRESSION): BOOLEAN
		do
			across node.mixed as c loop
				if attached {AST_NODE} c.item as expr then
					expr.accept (Current)
				else
					buffer.append (c.item.out)
				end
			end
		end

	visit_initializer (node: AST_INITIALIZER): BOOLEAN
		do
			buffer.append ("{ ")
			from
				node.elements.start
			until
				node.elements.after
			loop
				node.elements.item.accept (Current)
				if not node.elements.islast then
					buffer.append (", ")
				end
				node.elements.forth
			end
			buffer.append (" }")
		end

	visit_bit_field (node: AST_BITFIELD): BOOLEAN
		do
			node.type.accept (Current)
			if node.identifier /= Void then
				buffer.append (" ")
				buffer.append (node.orig_c_identifier)
			end
			buffer.append (" : ")
			node.size.accept (Current)
		end

	visit_switch_statement (node: AST_SWITCH_STATEMENT): BOOLEAN
		do
			buffer.append ("switch (")
			node.condition.accept (Current)
			buffer.append (")%N")
			print_indent
			node.statements.accept (Current)
		end

	visit_switch_case (node: AST_SWITCH_CASE): BOOLEAN
		do
			buffer.append ("case ")
			node.condition.accept (Current)
			buffer.append (":")
		end

	visit_switch_default (node: AST_SWITCH_DEFAULT): BOOLEAN
		do
			buffer.append ("default:")
		end

	visit_pointer_type (node: AST_POINTER_TYPE): BOOLEAN
		do
			if attached {AST_FUNCTION_TYPE} node.non_pointer_type then
				print_function_pointer_head (node)
				print_function_pointer_tail (node)
			elseif attached {AST_ARRAY_TYPE} node.non_pointer_type then
				print_array_pointer (node, Void)
			else
				node.type.accept (Current)
				buffer.append (" *")
				if node.is_const then
					buffer.append (" const")
				end
				if node.is_volatile then
					buffer.append (" volatile")
				end
			end
		end

	visit_basic_type (node: AST_BASIC_TYPE): BOOLEAN
		do
			inspect
				node.type
			when Ast_bt_void then
				buffer.append ("void")
			when Ast_bt_bool then
				buffer.append ("_Bool")
			when Ast_bt_char then
				buffer.append ("char")
			when Ast_bt_short then
				buffer.append ("short")
			when Ast_bt_int then
				buffer.append ("int")
			when Ast_bt_long then
				buffer.append ("long")
			when Ast_bt_long_long then
				buffer.append ("long long")
			when Ast_bt_unsigned_char then
				buffer.append ("unsigned char")
			when Ast_bt_unsigned_short then
				buffer.append ("unsigned short")
			when Ast_bt_unsigned_int then
				buffer.append ("unsigned int")
			when Ast_bt_unsigned_long then
				buffer.append ("unsigned long")
			when Ast_bt_unsigned_long_long then
				buffer.append ("unsigned long long")
			when Ast_bt_float then
				buffer.append ("float")
			when Ast_bt_double then
				buffer.append ("double")
			when Ast_bt_long_double then
				buffer.append ("long double")
			when Ast_bt_builtin_va_list then
				buffer.append ("__builtin_va_list")
			end
			if node.is_const then
				buffer.append (" const")
			end
			if node.is_volatile then
				buffer.append (" volatile")
			end
		end

	visit_eiffel_string_type (node: AST_EIFFEL_STRING_TYPE): BOOLEAN
		do
			buffer.append ("char *")
		end

	visit_eiffel_type (node: AST_EIFFEL_TYPE): BOOLEAN
		do
			buffer.append ("[" + node.type_name + "]")
		end

	visit_typedef_type (node: AST_TYPEDEF_TYPE): BOOLEAN
		do
			buffer.append (node.identifier)
			if node.is_const then
				buffer.append (" const")
			end
			if node.is_volatile then
				buffer.append (" volatile")
			end
		end

	visit_struct_type (node: AST_STRUCT_TYPE): BOOLEAN
		do
			if
				node.declaration_binding = Void or else
				not (
					node.declaration_binding.orig_c_identifier.starts_with ("__anon") and
					attached node.declaration_binding.typedef_name as typedef_name
				)
			then
				if node.is_union then
					buffer.append ("union ")
				else
					buffer.append ("struct ")
				end
				buffer.append (node.identifier)
				if node.is_const then
					buffer.append (" const")
				end
				if node.is_volatile then
					buffer.append (" volatile")
				end
			else
				buffer.append (typedef_name)
			end
		end

	visit_array_type (node: AST_ARRAY_TYPE): BOOLEAN
		local
			cur: AST_ARRAY_TYPE
		do
			from
				cur := node
			until
				cur = Void
			loop
				buffer.append ("[")
				if cur.orig_c_dimension /= Void then
					cur.orig_c_dimension.accept (Current)
				end
				buffer.append ("]")
				cur ?= cur.type
			end
		end

	visit_enum_type (node: AST_ENUM_TYPE): BOOLEAN
		do
			buffer.append ("enum ")
			buffer.append (node.identifier)
		end

	visit_boolean_constant (node: AST_BOOLEAN_CONSTANT): BOOLEAN
		do
			if node.value then
				buffer.append ("1")
			else
				buffer.append ("0")
			end
		end

	visit_eiffel_jump (node: AST_EIFFEL_JUMP): BOOLEAN
		do
			buffer.append ("[goto state ")
			if node.index /= Void then
				buffer.append (node.index.out)
			else
				buffer.append ("-1 (return)")
			end
			buffer.append ("]")
		end

	visit_eiffel_jump_target (node: AST_EIFFEL_JUMP_TARGET): BOOLEAN
		do
			buffer.append ("[state: ")
			buffer.append (node.index.out)
			if node.comment /= Void then
				buffer.append (", comment: ")
				buffer.append (node.comment)
			end
			buffer.append ("]")
		end

	visit_assembler_statement (node: AST_ASSEMBLER_STATEMENT): BOOLEAN
		do
			buffer.append ("__asm__ volatile (")
			across node.template as c loop
				if c.target_index > 1 then
					buffer.append (" ")
				end
				buffer.append (c.item)
			end
			buffer.append ("%N")

			print_indent
			buffer.append ("%T%T: ")
			across node.output_operands as c loop
				if c.target_index > 1 then
					buffer.append (", ")
				end
				c.item.accept (Current)
			end
			buffer.append ("%N")

			print_indent
			buffer.append ("%T%T: ")
			across node.input_operands as c loop
				if c.target_index > 1 then
					buffer.append (", ")
				end
				c.item.accept (Current)
			end
			buffer.append ("%N")

			if not node.clobbered_regs.is_empty then
				print_indent
				buffer.append ("%T%T: ")
				across node.clobbered_regs as c loop
					if c.target_index > 1 then
						buffer.append (", ")
					end
					buffer.append (c.item)
				end
				buffer.append ("%N")
			end

			print_indent
			buffer.append ("%T);")
		end

	visit_asm_operand (node: AST_ASM_OPERAND): BOOLEAN
		do
			buffer.append (node.constraint)
			buffer.append (" (")
			node.orig_c_expression.accept (Current)
			buffer.append (")")
		end

	visit_attribute (node: AST_GCC_ATTRIBUTE): BOOLEAN
		do
			buffer.append (node.identifier)
			if not node.expressions.is_empty then
				buffer.append ("(")
				across node.expressions as c loop
					if c.target_index > 1 then
						buffer.append (", ")
					end
					c.item.accept (Current)
				end
				buffer.append (")")
			end
		end

	visit_attribute_specifier (node: AST_GCC_ATTRIBUTE_SPECIFIER): BOOLEAN
		do
			buffer.append ("__attribute__((")
			across node.attributes as c loop
				if c.target_index > 1 then
					buffer.append (", ")
				end
				c.item.accept (Current)
			end
			buffer.append ("))")
		end

feature {ANY}

	visit_function_type (node: AST_FUNCTION_TYPE): BOOLEAN
		do
			buffer.append ("?AST_FUNCTION_TYPE?")
		end

feature {ANY}

	end_visit_array_type (node: AST_ARRAY_TYPE) do end
	end_visit_assembler_statement (node: AST_ASSEMBLER_STATEMENT) do end
	end_visit_asm_operand (node: AST_ASM_OPERAND) do end
	end_visit_attribute (node: AST_GCC_ATTRIBUTE) do end
	end_visit_attribute_specifier (node: AST_GCC_ATTRIBUTE_SPECIFIER) do end
	end_visit_basic_type (node: AST_BASIC_TYPE) do end
	end_visit_bit_field (node: AST_BITFIELD) do end
	end_visit_builtin_va_arg_call (node: AST_BUILTIN_VA_ARG_CALL) do end
	end_visit_alignof_call (node: AST_ALIGNOF_CALL) do end
	end_visit_initializer (node: AST_INITIALIZER) do end
	end_visit_assignment_expression (node: AST_ASSIGNMENT_EXPRESSION) do end
	end_visit_boolean_constant (node: AST_BOOLEAN_CONSTANT) do end
	end_visit_cast_expression (node: AST_CAST_EXPRESSION) do end
	end_visit_character_constant (node: AST_CHARACTER_CONSTANT) do end
	end_visit_compound_statement (node: AST_COMPOUND_STATEMENT) do end
	end_visit_decimal_constant (node: AST_DECIMAL_CONSTANT) do end
	end_visit_eiffel_expression (node: AST_EIFFEL_EXPRESSION) do end
	end_visit_eiffel_jump (node: AST_EIFFEL_JUMP) do end
	end_visit_eiffel_jump_target (node: AST_EIFFEL_JUMP_TARGET) do end
	end_visit_eiffel_string_type (node: AST_EIFFEL_STRING_TYPE) do end
	end_visit_eiffel_type (node: AST_EIFFEL_TYPE) do end
	end_visit_enum_declaration (node: AST_ENUM_DECLARATION) do end
	end_visit_enum_type (node: AST_ENUM_TYPE) do end
	end_visit_expression_statement (node: AST_EXPRESSION_STATEMENT) do end
	end_visit_floating_point_constant (node: AST_FLOATING_POINT_CONSTANT) do end
	end_visit_function_call (node: AST_FUNCTION_CALL) do end
	end_visit_function_declaration (node: AST_FUNCTION_DECLARATION) do end
	end_visit_function_type (node: AST_FUNCTION_TYPE) do end
	end_visit_generic_expression (node: AST_GENERIC_EXPRESSION) do end
	end_visit_hex_constant (node: AST_HEX_CONSTANT) do end
	end_visit_if_statement (node: AST_IF_STATEMENT) do end
	end_visit_jump_statement (node: AST_JUMP_STATEMENT) do end
	end_visit_labeled_statement (node: AST_LABELED_STATEMENT) do end
	end_visit_octal_constant (node: AST_OCTAL_CONSTANT) do end
	end_visit_parameter_list (node: AST_PARAMETER_LIST) do end
	end_visit_parenthesized_expression (node: AST_PARENTHESIZED_EXPRESSION) do end
	end_visit_pointer_type (node: AST_POINTER_TYPE) do end
	end_visit_postfix_expression (node: AST_POSTFIX_EXPRESSION) do end
	end_visit_sizeof_expression (node: AST_SIZEOF_EXPRESSION) do end
	end_visit_string_constant (node: AST_STRING_CONSTANT) do end
	end_visit_struct_declaration (node: AST_STRUCT_DECLARATION) do end
	end_visit_struct_type (node: AST_STRUCT_TYPE) do end
	end_visit_switch_case (node: AST_SWITCH_CASE) do end
	end_visit_switch_default (node: AST_SWITCH_DEFAULT) do end
	end_visit_switch_statement (node: AST_SWITCH_STATEMENT) do end
	end_visit_translation_unit (node: AST_TRANSLATION_UNIT) do end
	end_visit_typedef_declaration (node: AST_TYPEDEF_DECLARATION) do end
	end_visit_typedef_type (node: AST_TYPEDEF_TYPE) do end
	end_visit_unary_expression (node: AST_UNARY_EXPRESSION) do end
	end_visit_variable_declaration (node: AST_VARIABLE_DECLARATION) do end
	end_visit_variable_expression (node: AST_VARIABLE_EXPRESSION) do end
	end_visit_while_statement (node: AST_WHILE_STATEMENT) do end

	pre_visit (node: AST_NODE) do end
	post_visit (node: AST_NODE) do end

feature {NONE} -- helper routines

	print_storage_class_specifiers (l: LIST [INTEGER])
		require
			l /= Void
		do
			across l as c loop
				if c.target_index > 1 then
					buffer.append (" ")
				end
				inspect
					c.item
				when C_kw_extern then
					buffer.append ("extern")
				when C_kw_static then
					buffer.append ("static")
				when C_kw_auto then
					buffer.append ("auto")
				when C_kw_register then
					buffer.append ("register")
				when C_kw_inline then
					buffer.append ("__inline")
				end
			end
		end

	print_function_pointer_head (node: AST_TYPE)
		local
			cur: AST_POINTER_TYPE
		do
			if attached {AST_POINTER_TYPE} node as pt and then attached {AST_FUNCTION_TYPE} pt.non_pointer_type as ft then
				print_function_pointer_head (ft.return_type)
				buffer.append ("(")
				from
					cur := pt
				until
					cur = Void
				loop
					buffer.append (" *")
					if cur.is_const then
						buffer.append (" const")
					end
					if cur.is_volatile then
						buffer.append (" volatile")
					end
					cur ?= cur.type
				end
			else
				node.accept (Current)
			end
		end

	print_function_pointer_tail (node: AST_TYPE)
		do
			if attached {AST_POINTER_TYPE} node as pt and then attached {AST_FUNCTION_TYPE} pt.non_pointer_type as ft then
				buffer.append (")(")
				ft.parameter_list.accept (Current)
				buffer.append (")")
				print_function_pointer_tail (ft.return_type)
			end
		end

	print_array_pointer (node: AST_POINTER_TYPE; identifier: STRING)
		require
			node /= Void and then attached {AST_ARRAY_TYPE} node.non_pointer_type
		local
			array_type: AST_ARRAY_TYPE
			cur: AST_POINTER_TYPE
		do
			array_type ?= node.non_pointer_type
			array_type.type.accept (Current)
			buffer.append (" (")
			from
				cur := node
			until
				cur = Void
			loop
				buffer.append (" *")
				if cur.is_const then
					buffer.append (" const")
				end
				if cur.is_volatile then
					buffer.append (" volatile")
				end
				cur ?= cur.type
			end
			if identifier /= Void then
				buffer.append (identifier)
			end
			buffer.append (")")
			array_type.accept (Current)
		end

feature {ANY} -- indent handling

	indent: INTEGER_32 assign set_indent

	set_indent (a_indent: INTEGER_32)
		require
			a_indent >= 0
		do
			indent := a_indent
		ensure
			indent = a_indent
		end

feature {NONE} -- indent handling

	print_indent
		do
			buffer.append (multiply ("%T", indent))
		end

	multiply (s: STRING; count: INTEGER): STRING
		require
			count >= 0
		do
			if count = 0 then
				Result := ""
			else
				Result := s
				Result.multiply (count)
			end
		ensure
			Result /= Void
		end

end
