class
	AST_EIFFEL_CREATOR

inherit
	AST_VISITOR
	AST_UTILITIES
	NATIVE_FUNCTIONS_WRITER

create
	make

feature {NONE}

	make
		do
			create buffer.make_empty
			possible_semicolon_positions := create {TWO_WAY_LIST [INTEGER_32]}.make
		end

feature {ANY}

	buffer: STRING

feature {ANY}

	visit_enum_declaration (node: AST_ENUM_DECLARATION): BOOLEAN
		local
			l_eiffel_type: STRING_8
		do
			l_eiffel_type := node.replacement_type.eiffel_code

			buffer.append ("class%N")
			buffer.append ("	" + eiffel_class_name (node.identifier) + "%N%N")
			buffer.append ("feature {ANY}%N%N")
			across node.fields as c loop
				buffer.append ("	" + c.item.identifier + ": " + l_eiffel_type + " = ")
				c.item.initializer.accept (Current)
				buffer.append ("%N%N")
			end
			buffer.append ("end%N")
		end

	visit_variable_declaration (node: AST_VARIABLE_DECLARATION): BOOLEAN
		do
			if node.is_extern then
				write_native_variable (node, buffer)
			else
				buffer.append (node.identifier + ": ")
				node.type.accept (Current)
			end
		end

	visit_function_declaration (node: AST_FUNCTION_DECLARATION): BOOLEAN
		do
			if node.is_regular then
				buffer.append ("	")
				buffer.append (node.identifier)
				if not node.parameter_list.parameters.is_empty then
					buffer.append (" (")
					node.parameter_list.accept (Current)
					buffer.append (")")
				end
				if not attached {AST_BASIC_TYPE} node.return_type as basic_type or else basic_type.type /= Ast_bt_void then
					buffer.append (": ")
					node.return_type.accept (Current)
				end
				buffer.append ("%N")

				if not node.oo_preconditions.is_empty then
					buffer.append ("		require%N")
					across node.oo_preconditions as c loop
						buffer.append ("			")
						c.item.accept (Current)
						buffer.append ("%N")
					end
				end

				indent := indent + 2
				if
					not node.locals.is_empty or
					node.has_jump_statements or
					node.needs_alloca_factory or
					node.jump_variables /= Void
				then
					print_indent
					buffer.append ("local%N")
					indent := indent + 1
					across node.locals as c loop
						print_indent
						c.item.accept (Current)
						buffer.append ("%N")
					end
					if node.has_jump_statements then
						buffer.append ("			l_pos: INTEGER_32%N")
					end
					if node.needs_alloca_factory then
						buffer.append ("			l_aaf: CE_ALLOCA_FACTORY%N")
					end
					if node.jump_variables /= Void then
						buffer.append ("			" + node.jump_variables + ": BOOLEAN%N")
					end
					indent := indent - 1
				end
				print_indent
				buffer.append ("do%N")
				indent := indent + 1
				initialize_pointers (node)
				indent := indent - 1
				if node.has_jump_statements then
					buffer.append ("			from until l_pos = -1 loop -- -1: return%N")
					buffer.append ("				inspect l_pos%N%N")
					buffer.append ("				-- <function body>%N")
					indent := indent + 2
				end
				node.statements.accept (Current)
				if node.has_jump_statements then
					buffer.append ("				end%N")
					buffer.append ("				-- </function body>%N%N")
					buffer.append ("				l_pos := l_pos + 1%N")
					buffer.append ("			end%N")
					indent := indent - 2
				end
				if not node.oo_postconditions.is_empty then
					buffer.append ("		ensure%N")
					across node.oo_postconditions as c loop
						buffer.append ("			")
						c.item.accept (Current)
						buffer.append ("%N")
					end
				end
				if node.is_longjmp_target then
					buffer.append ("		rescue%N")
					buffer.append ("			if attached {CE_EXCEPTION} (create {EXCEPTION_MANAGER}).last_exception as ex then%N")
					buffer.append ("				if ex.routine = $" + node.identifier + " then%N")
					buffer.append ("					l_pos := ex.position + 1%N")
					buffer.append ("					retry%N")
					buffer.append ("				end%N")
					buffer.append ("				ex.re_raise%N")
					buffer.append ("			end%N")
				end
				print_indent
				buffer.append ("end")
				indent := indent - 2
			elseif node.is_extern then
				write_native_function (node, buffer)
			else
				(create {EXCEPTIONS}).raise ("?AST_FUNCTION_DECLARATION?")
			end
		end

	visit_compound_statement (node: AST_COMPOUND_STATEMENT): BOOLEAN
		local
			l_old_semicolon_count: INTEGER_32
		do
			indent := indent + 1
			across node.statements as c loop
				l_old_semicolon_count := possible_semicolon_positions.count

				if not attached {AST_EIFFEL_JUMP_TARGET} c.item then
					print_indent
				end
				c.item.accept (Current)
				buffer.append ("%N")

				if
					l_old_semicolon_count = possible_semicolon_positions.count and
					not attached {AST_IF_STATEMENT} c.item and
					not attached {AST_WHILE_STATEMENT} c.item
				then
					possible_semicolon_positions.extend (buffer.count)
				end
			end
			indent := indent - 1
			handle_semicolons
		end

	visit_expression_statement (node: AST_EXPRESSION_STATEMENT): BOOLEAN
		do
			if node.expression /= Void then
				node.expression.accept (Current)
			end
		end

	visit_hex_constant (node: AST_HEX_CONSTANT): BOOLEAN
		do
			buffer.append (node.value)
		end

	visit_octal_constant (node: AST_OCTAL_CONSTANT): BOOLEAN
		do
			buffer.append (node.value)
		end

	visit_decimal_constant (node: AST_DECIMAL_CONSTANT): BOOLEAN
		local
			l_count: INTEGER_32
		do
			inspect
				node.type_binding.type
			when Ast_bt_int, Ast_bt_long then
				-- that's ok
			when Ast_bt_long_long then
				buffer.append ("{INTEGER_64} ")
			when Ast_bt_unsigned_int, Ast_bt_unsigned_long then
				buffer.append ("{NATURAL_32} ")
			when Ast_bt_unsigned_long_long then
				buffer.append ("{NATURAL_64} ")
			end

			l_count := node.value.count
			across node.value as c loop
				if c.target_index > 1 and (l_count - c.target_index + 1) \\ 3 = 0 then
					buffer.append_character ('_')
				end
				buffer.append_character (c.item)
			end
		end

	visit_character_constant (node: AST_CHARACTER_CONSTANT): BOOLEAN
		do
			if node.is_eiffel_character then
				buffer.append ("'")
				buffer.append (node.value)
				buffer.append ("'")
			else
				buffer.append ("('")
				buffer.append (node.value)
				buffer.append ("').code")
			end
		end

	visit_string_constant (node: AST_STRING_CONSTANT): BOOLEAN
		do
			if node.is_eiffel_string then
				buffer.append ("%"")
				buffer.append (node.value)
				buffer.append ("%"")
			else
				buffer.append ("ce_cstring (%"")
				buffer.append (node.value)
				buffer.append ("%")")
			end
		end

	visit_floating_point_constant (node: AST_FLOATING_POINT_CONSTANT): BOOLEAN
		do
			inspect
				node.type_binding.type
			when Ast_bt_float then
				buffer.append ("{REAL_32} ")
				buffer.append (node.value)
			when Ast_bt_double then
				-- no prefix
				buffer.append (node.value)
			when Ast_bt_long_double then
				buffer.append ("real_96_from_real_64 (")
				buffer.append (node.value)
				buffer.append (")")
			end
		end

	visit_postfix_expression (node: AST_POSTFIX_EXPRESSION): BOOLEAN
		local
			b: BOOLEAN
		do
			inspect
				node.type
			when C_op_open_square_bracket then
				b := attached {AST_POSTFIX_EXPRESSION} node.parent as tmp and then tmp.type = C_op_open_square_bracket
				if b then
					buffer.append ("(")
				end
				node.main_expression.accept (Current)
				buffer.append ("[")
				node.array_expression.accept (Current)
				buffer.append ("]")
				if b then
					buffer.append (")")
				end
			when C_op_point, C_op_arrow then
				node.main_expression.accept (Current)
				if node.type = C_op_arrow then
					buffer.append (".item")
				end
				buffer.append (".")
				if attached {AST_VARIABLE_DECLARATION} node.declaration_binding as tmp then
					buffer.append (tmp.identifier)
				elseif attached {AST_BITFIELD} node.declaration_binding as tmp then
					buffer.append (tmp.identifier)
				else
					buffer.append ("?NO_DECLARATION?")
				end
			end
		end

	visit_parenthesized_expression (node: AST_PARENTHESIZED_EXPRESSION): BOOLEAN
		do
			if
				(attached {AST_PARENTHESIZED_EXPRESSION} node.expression) or -- `((foo))' -> `(foo)'
				(attached {AST_VARIABLE_EXPRESSION} node.expression) or -- `(i)' -> `i'
				(attached {AST_DECIMAL_CONSTANT} node.expression and attached {AST_ASSIGNMENT_EXPRESSION} node.non_parentheses_parent) -- `foo := (123)' -> `foo := 123'
			then
				node.expression.accept (Current)
			else
				buffer.append ("(")
				node.expression.accept (Current)
				buffer.append (")")
			end
		end

	visit_variable_expression (node: AST_VARIABLE_EXPRESSION): BOOLEAN
		local
			l_declaration: AST_EIFFEL_FEATURE
		do
			-- use the name of the declaration; it might have been renamed
			l_declaration := node.declaration_binding
			if attached {AST_VARIABLE_DECLARATION} l_declaration then
				if
					attached {AST_PARAMETER_LIST} l_declaration.parent as pl and then
					pl.is_vararg
				then
					buffer.append ("a_args.")
				end
				buffer.append (l_declaration.identifier)
				if
					not l_declaration.is_extern and
					attached {AST_TRANSLATION_UNIT} l_declaration.parent
				then
					buffer.append (".item")
				end
			else
				buffer.append (l_declaration.identifier)
			end
		end

	visit_parameter_list (node: AST_PARAMETER_LIST): BOOLEAN
		local
			l_array: ARRAY [TUPLE [identifier: STRING_8; type: AST_TYPE]]
		do
			create l_array.make_filled (Void, 1, node.parameters.count)
			across node.parameters as c loop
				if attached {AST_VARIABLE_DECLARATION} c.item as vd then
					l_array[c.target_index] := [vd.identifier, vd.type]
				elseif attached {AST_TYPE} c.item as t then
					l_array[c.target_index] := ["a_arg" + c.target_index.out, t]
				end
			end

			if node.is_vararg then
				buffer.append ("a_args: TUPLE [")
			end
			across l_array as c loop
				if c.target_index > 1 then
					if c.item.type.is_equal_type (l_array[c.target_index - 1].type) then
						buffer.append (", ")
					else
						buffer.append (": ")
						l_array[c.target_index - 1].type.accept (Current)
						buffer.append ("; ")
					end
				end
				buffer.append (c.item.identifier)
				if c.target_index = l_array.count then
					buffer.append (": ")
					c.item.type.accept (Current)
				end
			end
			if node.is_vararg then
				buffer.append ("]")
			end
		end

	visit_if_statement (node: AST_IF_STATEMENT): BOOLEAN
		local
			omit_end: BOOLEAN
		do
			buffer.append ("if ")
			node.condition.accept (Current)
			buffer.append (" then%N")
			node.then_statements.accept (Current)
			if node.else_statements /= Void then
				print_indent
				if
					node.else_statements.statements.count = 1 and then
					attached {AST_IF_STATEMENT} node.else_statements.statements.first as if_statement
				then
					buffer.append ("else")
					if_statement.accept (Current)
					omit_end := True
				else
					buffer.append ("else%N")
					node.else_statements.accept (Current)
				end
			end
			if not omit_end then
				print_indent
				buffer.append ("end")
			end
		end

	visit_generic_expression (node: AST_GENERIC_EXPRESSION): BOOLEAN
		do
			inspect
				node.type
			when C_op_plus then
				node.left.accept (Current)
				buffer.append (" + ")
				node.right.accept (Current)
			when C_op_minus then
				node.left.accept (Current)
				buffer.append (" - ")
				node.right.accept (Current)
			when C_op_mult then
				node.left.accept (Current)
				buffer.append (" * ")
				node.right.accept (Current)
			when C_op_div then
				node.left.accept (Current)
				if
					attached {AST_BASIC_TYPE} node.left.type_binding as tmp and then
					(tmp.type = Ast_bt_float or tmp.type = Ast_bt_double or tmp.type = Ast_bt_long_double)
				then
					buffer.append (" / ")
				else
					buffer.append (" // ")
				end
				node.right.accept (Current)
			when C_op_modulo then
				node.left.accept (Current)
				if
					attached {AST_BASIC_TYPE} node.left.type_binding as tmp and then
					(tmp.type = Ast_bt_float or tmp.type = Ast_bt_double or tmp.type = Ast_bt_long_double)
				then
					buffer.append (" %% ")
				else
					buffer.append (" \\ ")
				end
				node.right.accept (Current)
			when C_op_logical_and then
				node.left.accept (Current)
				buffer.append (" and then ")
				node.right.accept (Current)
			when C_op_logical_or then
				node.left.accept (Current)
				buffer.append (" or else ")
				node.right.accept (Current)
			when C_op_smaller then
				node.left.accept (Current)
				buffer.append (" < ")
				node.right.accept (Current)
			when C_op_greater then
				node.left.accept (Current)
				buffer.append (" > ")
				node.right.accept (Current)
			when C_op_smaller_equal then
				node.left.accept (Current)
				buffer.append (" <= ")
				node.right.accept (Current)
			when C_op_greater_equal then
				node.left.accept (Current)
				buffer.append (" >= ")
				node.right.accept (Current)
			when C_op_equal then
				node.left.accept (Current)
				buffer.append (" = ")
				node.right.accept (Current)
			when C_op_not_equal then
				node.left.accept (Current)
				buffer.append (" /= ")
				node.right.accept (Current)
			when C_op_bitand then
				-- don't use `&'; different operator precedence
				buffer.append ("(")
				node.left.accept (Current)
				buffer.append (").bit_and (")
				node.right.accept (Current)
				buffer.append (")")
			when C_op_bitor then
				-- don't use `|'; different operator precedence
				buffer.append ("(")
				node.left.accept (Current)
				buffer.append (").bit_or (")
				node.right.accept (Current)
				buffer.append (")")
			when C_op_left_shift then
				buffer.append ("(")
				node.left.accept (Current)
				buffer.append (").bit_shift_left (")
				node.right.accept (Current)
				buffer.append (")")
			when C_op_right_shift then
				buffer.append ("(")
				node.left.accept (Current)
				buffer.append (").bit_shift_right (")
				node.right.accept (Current)
				buffer.append (")")
			when C_op_exor then
				buffer.append ("(")
				node.left.accept (Current)
				buffer.append (").bit_xor (")
				node.right.accept (Current)
				buffer.append (")")
			end
		end

	visit_assignment_expression (node: AST_ASSIGNMENT_EXPRESSION): BOOLEAN
		do
			inspect
				node.type
			when C_op_assign then
				if
					attached {AST_VARIABLE_EXPRESSION} node.target as ve and then
					attached {AST_VARIABLE_DECLARATION} ve.declaration_binding as vd and then
					vd.is_extern -- assignment to an external variable?
				then
					-- prefix with "Current." so that the associated assigner command is called
					buffer.append ("Current.")
				end
				node.target.accept (Current)
				buffer.append (" := ")
				node.source.accept (Current)
			else
				buffer.append ("?AST_ASSIGNMENT_EXPRESSION?")
			end
		end

	visit_unary_expression (node: AST_UNARY_EXPRESSION): BOOLEAN
		local
			expr: AST_EXPRESSION
		do
			inspect
				node.type
			when C_op_increment then
				buffer.append ("++")
				node.expression.accept (Current)
			when C_op_decrement then
				buffer.append ("--")
				node.expression.accept (Current)
			when C_op_bitand then
				buffer.append ("&")
				node.expression.accept (Current)
			when C_op_mult then
				node.expression.accept (Current)
				buffer.append (".item")
			when C_op_plus then
				buffer.append ("+")
				node.expression.accept (Current)
			when C_op_minus then
				buffer.append ("-")
				node.expression.accept (Current)
			when C_op_neg then
				buffer.append ("(")
				node.expression.accept (Current)
				buffer.append (").bit_not")
			when C_op_not then
				expr := create_negated_expression (node.expression)
				if attached {AST_UNARY_EXPRESSION} expr as tmp and then tmp.type = C_op_not then
					buffer.append ("not ")
					node.expression.accept (Current)
				else
					expr.parent := node -- needs to be in the tree, at least for CLASS_UTILITIES.get_class_prefix
					expr.accept (Current)
				end
			end
		end

	visit_while_statement (node: AST_WHILE_STATEMENT): BOOLEAN
		local
			negated_condition: AST_EXPRESSION
		do
			buffer.append ("from%N")
			print_indent
			buffer.append ("until%N")
			indent := indent + 1
			print_indent
			indent := indent - 1
			negated_condition := create_negated_expression (node.condition)
			negated_condition.parent := node -- needs to be in the tree, at least for CLASS_UTILITIES.get_class_prefix
			negated_condition.accept (Current)
			buffer.append ("%N")
			print_indent
			buffer.append ("loop%N")
			node.statements.accept (Current)
			print_indent
			buffer.append ("end")
		end

	visit_function_type (node: AST_FUNCTION_TYPE): BOOLEAN
		do
			buffer.append ("?AST_FUNCTION_TYPE?")
		end

	visit_function_call (node: AST_FUNCTION_CALL): BOOLEAN
		local
			has_make_cast: BOOLEAN
			identifier: STRING
			vararg: BOOLEAN
		do
			if attached {AST_FUNCTION_DECLARATION} node.declaration_binding as tmp then
				identifier := tmp.identifier
				vararg := tmp.parameter_list.is_vararg
			else
				identifier := ({AST_UNARY_EXPRESSION} #? node.identifier_expr).expression.eiffel_code
				if attached {AST_EXPRESSION_STATEMENT} node.parent then
					identifier.append (".call")
				else
					if attached {AST_BASIC_TYPE} node.type_binding then
						identifier.append ("." + node.type_binding.eiffel_code.as_lower + "_item")
					elseif attached {AST_POINTER_TYPE} node.type_binding then
						has_make_cast := True
						identifier.prepend ("create {" + effective_type (node.type_binding) + "}.make_cast (")
						identifier.append (".pointer_item")
					elseif attached {AST_STRUCT_TYPE} node.type_binding then
						identifier.prepend ("{" + node.type_binding.eiffel_code + "} #? ")
						identifier.append (".ce_class_item")
					else
						buffer.append ("[FUNCTION_CALL to " + node.type_binding.generator + "]")
					end
				end
				vararg := True
			end

			buffer.append (identifier)
			if not node.parameters.is_empty or vararg then
				buffer.append (" (")
				if vararg then
					buffer.append ("[")
				end
				across node.parameters as c loop
					if c.target_index > 1 then
						buffer.append (", ")
					end
					c.item.accept (Current)
				end
				if vararg then
					buffer.append ("]")
				end
				buffer.append (")")
			end
			if has_make_cast then
				buffer.append (")")
			end
		end

	visit_eiffel_expression (node: AST_EIFFEL_EXPRESSION): BOOLEAN
		do
			across node.mixed as c loop
				if attached {AST_NODE} c.item as expr then
					expr.accept (Current)
				elseif attached {STRING_8} c.item as s then
					if
						c.target_index = node.mixed.count and then
						s.starts_with (" -- ")
					then
						possible_semicolon_positions.extend (buffer.count + 1)
					end
					buffer.append (s)
				end
			end
		end

	visit_boolean_constant (node: AST_BOOLEAN_CONSTANT): BOOLEAN
		do
			if node.value then
				buffer.append ("True")
			else
				buffer.append ("False")
			end
		end

	visit_eiffel_jump (node: AST_EIFFEL_JUMP): BOOLEAN
		do
			buffer.append ("l_pos := -1 + ")
			if node.index = Void then
				buffer.append ("-1")
			else
				buffer.append_natural_32 (node.index)
			end
			if node.comment /= Void then
				buffer.append (" -- ")
				buffer.append (node.comment)
			end
		end

	visit_eiffel_jump_target (node: AST_EIFFEL_JUMP_TARGET): BOOLEAN
		do
			buffer.append ("				when " + node.index.out + " then")
			if node.comment /= Void then
				buffer.append (" -- ")
				buffer.append (node.comment)
			end
			indent := indent + node.indent_update
		end

	visit_basic_type (node: AST_BASIC_TYPE): BOOLEAN
		do
			inspect
				node.type
			when Ast_bt_void then
				buffer.append ("ANY")
			when Ast_bt_bool then
				buffer.append ("BOOLEAN")
			when Ast_bt_char then
				buffer.append ("INTEGER_8")
			when Ast_bt_short then
				buffer.append ("INTEGER_16")
			when Ast_bt_int, Ast_bt_long then
				buffer.append ("INTEGER_32")
			when Ast_bt_long_long then
				buffer.append ("INTEGER_64")
			when Ast_bt_unsigned_char then
				buffer.append ("NATURAL_8")
			when Ast_bt_unsigned_short then
				buffer.append ("NATURAL_16")
			when Ast_bt_unsigned_int, Ast_bt_unsigned_long then
				buffer.append ("NATURAL_32")
			when Ast_bt_unsigned_long_long then
				buffer.append ("NATURAL_64")
			when Ast_bt_float then
				buffer.append ("REAL_32")
			when Ast_bt_double then
				buffer.append ("REAL_64")
			when Ast_bt_long_double then
				buffer.append ("REAL_96")
			when Ast_bt_builtin_va_list then
				buffer.append ("CE_VA_LIST")
			end
		end

	visit_eiffel_string_type (node: AST_EIFFEL_STRING_TYPE): BOOLEAN
		do
			buffer.append ("STRING_8")
		end

	visit_eiffel_type (node: AST_EIFFEL_TYPE): BOOLEAN
		do
			buffer.append (node.type_name)
		end

	visit_array_type (node: AST_ARRAY_TYPE): BOOLEAN
		do
			buffer.append ("CE_ARRAY [")
			node.type.accept (Current)
			buffer.append ("]")
		end

	visit_pointer_type (node: AST_POINTER_TYPE): BOOLEAN
		do
			if attached {AST_FUNCTION_TYPE} node.type then
				buffer.append ("CE_ROUTINE")
			elseif buffer.ends_with ("create {") then
				-- from a cast; use effective type
				buffer.append (effective_type (node))
			else
				buffer.append ("CE_POINTER [")
				node.type.accept (Current)
				buffer.append ("]")
			end
		end

	visit_struct_type (node: AST_STRUCT_TYPE): BOOLEAN
		do
			buffer.append (eiffel_class_name (node.declaration_binding.identifier))
		end

	visit_enum_type (node: AST_ENUM_TYPE): BOOLEAN
		do
			buffer.append (eiffel_class_name (node.declaration_binding.identifier))
		end

	visit_alignof_call (node: AST_ALIGNOF_CALL): BOOLEAN
		do
			buffer.append ("ce_alignof ({")
			node.type.accept (Current)
			buffer.append ("})")
		end

	visit_initializer (node: AST_INITIALIZER): BOOLEAN
		local
			l_pos: INTEGER_32
			l_fields: TWO_WAY_LIST [AST_TYPE]
			l_cur: AST_ARRAY_TYPE
		do
			if attached {AST_ARRAY_TYPE} node.type_binding as at then
				buffer.append ("create {" + effective_type (at) + "}.make_initialized (<<%N")
				indent := indent + 2
				if attached {AST_BASIC_TYPE} at.type then
					-- newlines after 100 characters
					if not node.elements.is_empty then
						l_pos := buffer.count
						print_indent
						across node.elements as c loop
							c.item.accept (Current)
							if c.target_index < node.elements.count then
								buffer.append (",")
								if buffer.count - l_pos > 100 then
									buffer.append ("%N")
									l_pos := buffer.count
									print_indent
								else
									buffer.append (" ")
								end
							end
						end
						buffer.append ("%N")
					end
				else
					-- newlines after every element
					across node.elements as c loop
						print_indent
						c.item.accept (Current)
						if c.target_index < node.elements.count then
							buffer.append (",")
						end
						buffer.append ("%N")
					end
				end
				indent := indent - 1
				print_indent
				buffer.append (">>)")
				indent := indent - 1
			elseif
				attached {AST_STRUCT_TYPE} node.type_binding as st and then
				attached st.declaration_binding as sd
			then
				create l_fields.make
				across sd.fields as c loop
					if c.item.identifier /= Void then
						l_fields.extend (c.item.type)
					end
				end
				l_fields.start

				buffer.append ("create {" + st.eiffel_code + "}.make_initialized ([")
				across node.elements as c loop
					if c.target_index > 1 then
						buffer.append (", ")
					end
					c.item.accept (Current)
					l_fields.remove
				end
				across l_fields as c loop
					if not node.elements.is_empty or c.target_index > 1 then
						buffer.append (", ")
					end
					if attached {AST_BASIC_TYPE} c.item as bt then
						inspect
							bt.type
						when
							Ast_bt_char, Ast_bt_short, Ast_bt_int, Ast_bt_long, Ast_bt_long_long,
							Ast_bt_unsigned_char, Ast_bt_unsigned_short, Ast_bt_unsigned_int, Ast_bt_unsigned_long, Ast_bt_unsigned_long_long
						then
							buffer.append ("{" + c.item.eiffel_code + "} 0")
						when Ast_bt_float, Ast_bt_double then
							buffer.append ("{" + c.item.eiffel_code + "} 0.0")
						end
					elseif attached {AST_POINTER_TYPE} c.item as pt then
						buffer.append ("create {" + effective_type (pt) + "}.make_shared (default_pointer)")
					elseif attached {AST_ARRAY_TYPE} c.item as at then
						buffer.append ("create {" + effective_type (at) + "}.make (<<")
						from
							l_cur := at
						until
							l_cur = Void
						loop
							if l_cur /= at then
								buffer.append (", ")
							end
							l_cur.dimension.accept (Current)
							l_cur ?= l_cur.type
						end
						buffer.append (">>)")
					elseif attached {AST_STRUCT_TYPE} c.item as st2 then
						buffer.append ("create {" + st2.eiffel_code + "}.make")
					else
						(create {EXCEPTIONS}).raise (c.item.generator)
					end
				end
				buffer.append ("])")
			else
				(create {EXCEPTIONS}).raise (node.type_binding.eiffel_code)
			end
		end

feature {ANY}

	pre_visit (node: AST_NODE) do end
	post_visit (node: AST_NODE) do end

	visit_struct_declaration (node: AST_STRUCT_DECLARATION): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_jump_statement (node: AST_JUMP_STATEMENT): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_attribute (node: AST_GCC_ATTRIBUTE): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_attribute_specifier (node: AST_GCC_ATTRIBUTE_SPECIFIER): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_sizeof_expression (node: AST_SIZEOF_EXPRESSION): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_labeled_statement (node: AST_LABELED_STATEMENT): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_cast_expression (node: AST_CAST_EXPRESSION): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_translation_unit (node: AST_TRANSLATION_UNIT): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_bit_field (node: AST_BITFIELD): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_builtin_va_arg_call (node: AST_BUILTIN_VA_ARG_CALL): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_switch_case (node: AST_SWITCH_CASE): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_switch_default (node: AST_SWITCH_DEFAULT): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_switch_statement (node: AST_SWITCH_STATEMENT): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_typedef_declaration (node: AST_TYPEDEF_DECLARATION): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_typedef_type (node: AST_TYPEDEF_TYPE): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_assembler_statement (node: AST_ASSEMBLER_STATEMENT): BOOLEAN do buffer.append ("?" + node.generator + "?") end
	visit_asm_operand (node: AST_ASM_OPERAND): BOOLEAN do buffer.append ("?" + node.generator + "?") end

	end_visit_array_type (node: AST_ARRAY_TYPE) do end
	end_visit_assembler_statement (node: AST_ASSEMBLER_STATEMENT) do end
	end_visit_asm_operand (node: AST_ASM_OPERAND) do end
	end_visit_attribute (node: AST_GCC_ATTRIBUTE) do end
	end_visit_attribute_specifier (node: AST_GCC_ATTRIBUTE_SPECIFIER) do end
	end_visit_basic_type (node: AST_BASIC_TYPE) do end
	end_visit_bit_field (node: AST_BITFIELD) do end
	end_visit_builtin_va_arg_call (node: AST_BUILTIN_VA_ARG_CALL) do end
	end_visit_alignof_call (node: AST_ALIGNOF_CALL) do end
	end_visit_initializer (node: AST_INITIALIZER) do end
	end_visit_assignment_expression (node: AST_ASSIGNMENT_EXPRESSION) do end
	end_visit_boolean_constant (node: AST_BOOLEAN_CONSTANT) do end
	end_visit_cast_expression (node: AST_CAST_EXPRESSION) do end
	end_visit_character_constant (node: AST_CHARACTER_CONSTANT) do end
	end_visit_compound_statement (node: AST_COMPOUND_STATEMENT) do end
	end_visit_decimal_constant (node: AST_DECIMAL_CONSTANT) do end
	end_visit_eiffel_expression (node: AST_EIFFEL_EXPRESSION) do end
	end_visit_eiffel_jump (node: AST_EIFFEL_JUMP) do end
	end_visit_eiffel_jump_target (node: AST_EIFFEL_JUMP_TARGET) do end
	end_visit_eiffel_string_type (node: AST_EIFFEL_STRING_TYPE) do end
	end_visit_eiffel_type (node: AST_EIFFEL_TYPE) do end
	end_visit_enum_declaration (node: AST_ENUM_DECLARATION) do end
	end_visit_enum_type (node: AST_ENUM_TYPE) do end
	end_visit_expression_statement (node: AST_EXPRESSION_STATEMENT) do end
	end_visit_floating_point_constant (node: AST_FLOATING_POINT_CONSTANT) do end
	end_visit_function_call (node: AST_FUNCTION_CALL) do end
	end_visit_function_declaration (node: AST_FUNCTION_DECLARATION) do end
	end_visit_function_type (node: AST_FUNCTION_TYPE) do end
	end_visit_generic_expression (node: AST_GENERIC_EXPRESSION) do end
	end_visit_hex_constant (node: AST_HEX_CONSTANT) do end
	end_visit_if_statement (node: AST_IF_STATEMENT) do end
	end_visit_jump_statement (node: AST_JUMP_STATEMENT) do end
	end_visit_labeled_statement (node: AST_LABELED_STATEMENT) do end
	end_visit_octal_constant (node: AST_OCTAL_CONSTANT) do end
	end_visit_parameter_list (node: AST_PARAMETER_LIST) do end
	end_visit_parenthesized_expression (node: AST_PARENTHESIZED_EXPRESSION) do end
	end_visit_pointer_type (node: AST_POINTER_TYPE) do end
	end_visit_postfix_expression (node: AST_POSTFIX_EXPRESSION) do end
	end_visit_sizeof_expression (node: AST_SIZEOF_EXPRESSION) do end
	end_visit_string_constant (node: AST_STRING_CONSTANT) do end
	end_visit_struct_declaration (node: AST_STRUCT_DECLARATION) do end
	end_visit_struct_type (node: AST_STRUCT_TYPE) do end
	end_visit_switch_case (node: AST_SWITCH_CASE) do end
	end_visit_switch_default (node: AST_SWITCH_DEFAULT) do end
	end_visit_switch_statement (node: AST_SWITCH_STATEMENT) do end
	end_visit_translation_unit (node: AST_TRANSLATION_UNIT) do end
	end_visit_typedef_declaration (node: AST_TYPEDEF_DECLARATION) do end
	end_visit_typedef_type (node: AST_TYPEDEF_TYPE) do end
	end_visit_unary_expression (node: AST_UNARY_EXPRESSION) do end
	end_visit_variable_declaration (node: AST_VARIABLE_DECLARATION) do end
	end_visit_variable_expression (node: AST_VARIABLE_EXPRESSION) do end
	end_visit_while_statement (node: AST_WHILE_STATEMENT) do end

feature {NONE} -- semicolon handling

	possible_semicolon_positions: LIST [INTEGER_32]

	handle_semicolons
		local
			l_i: INTEGER_32
			l_in_comment: BOOLEAN
		do
			across possible_semicolon_positions.new_cursor.reversed as c loop
				from
					l_i := c.item
				until
					not l_in_comment and
					(
						l_i > buffer.count or else
						(buffer[l_i] /= '%T' and buffer[l_i] /= '%N' and buffer[l_i] /= '%R' and buffer[l_i] /= ' ' and buffer[l_i] /= '-')
					)
				loop
					if buffer[l_i] = '-' then
						l_in_comment := True
					elseif buffer[l_i] = '%N' then
						l_in_comment := False
					end
					l_i := l_i + 1
				end
				if l_i <= buffer.count and then buffer[l_i] = '(' then
					-- Nodes that emit comments have to add possible semicolon positions before the comment.
					-- Otherwise a required semicolon might end up in the comment.
					buffer.insert_character (';', c.item)
				end
			end
			possible_semicolon_positions.wipe_out
		end

feature {ANY} -- indent handling

	indent: INTEGER_32 assign set_indent

	set_indent (a_indent: INTEGER_32)
		require
			a_indent >= 0
		do
			indent := a_indent
		ensure
			indent = a_indent
		end

feature {NONE} -- helpers

	print_indent
		do
			buffer.append (multiply ("%T", indent))
		end

	initialize_pointers (a_node: AST_FUNCTION_DECLARATION)
		require
			a_node /= Void
		local
			printed: BOOLEAN
			cur: AST_ARRAY_TYPE
		do
			across a_node.locals as c loop
				if
					(
						(
							attached {AST_POINTER_TYPE} c.item.type or
							attached {AST_STRUCT_TYPE} c.item.type or
							(attached {AST_BASIC_TYPE} c.item.type as bt and then bt.type = Ast_bt_builtin_va_list)
						) and
						needs_initialization (a_node, c.item)
					) or
					attached {AST_ARRAY_TYPE} c.item.type
				then
					print_indent
					buffer.append ("create ")
					if
						(attached {AST_POINTER_TYPE} c.item.type as pt and then not attached {AST_FUNCTION_TYPE} pt.type) or
						attached {AST_ARRAY_TYPE} c.item.type
					then
						buffer.append ("{" + effective_type (c.item.type) + "} ")
					end
					buffer.append (c.item.identifier + ".make")
					if attached {AST_ARRAY_TYPE} c.item.type as array_type then
						buffer.append (" (<<")
						from
							cur := array_type
						until
							cur = Void
						loop
							if cur /= array_type then
								buffer.append (", ")
							end
							if cur.dimension /= Void then
								cur.dimension.accept (Current)
							else
								buffer.append ("TODO") -- not possible?!
							end
							cur ?= cur.type
						end
						buffer.append (">>)")
					elseif attached {AST_BASIC_TYPE} c.item.type as bt and then bt.type = Ast_bt_builtin_va_list then
						buffer.append ("_shared (default_pointer)")
					end
					buffer.append ("%N")
					printed := True
				end
			end
			if printed then
				possible_semicolon_positions.extend (buffer.count)
				buffer.append ("%N")
			end
		end

	needs_initialization (a_function: AST_FUNCTION_DECLARATION; a_node: AST_VARIABLE_DECLARATION): BOOLEAN
		do
			Result := needs_initialization_rec (
					a_node,
					a_function.cfg,
					create {HASH_SET [CFG_BASIC_BLOCK]}.make (16)
				)
		end

	needs_initialization_rec (a_node: AST_VARIABLE_DECLARATION; a_block: CFG_BASIC_BLOCK; a_visited: HASH_SET [CFG_BASIC_BLOCK]): BOOLEAN
		local
			l_status: CELL [NATURAL_8] -- 0: no usage found, 1: assigned to first, 2: read first
			l_visitor: AST_AGENT_VISITOR
		do
			if not a_visited.has (a_block) then
				a_visited.extend (a_block)

				create l_status.put (0)
				create l_visitor
				l_visitor.agents.visit_assignment_expression := agent (ag_node: AST_ASSIGNMENT_EXPRESSION; ag_visitor: AST_VISITOR): BOOLEAN
					do
						-- visit source first
						ag_node.source.accept (ag_visitor)
						ag_node.target.accept (ag_visitor)
					end (?, l_visitor)
				l_visitor.agents.end_visit_variable_expression := agent (ag_node: AST_VARIABLE_EXPRESSION; ag_decl: AST_VARIABLE_DECLARATION; ag_status: CELL [NATURAL_8])
					do
						if ag_status.item = 0 and ag_node.declaration_binding = ag_decl then
							if attached {AST_ASSIGNMENT_EXPRESSION} ag_node.parent as ae and then ae.target = ag_node then
								ag_status.put (1)
							else
								ag_status.put (2)
							end
						end
					end (?, a_node, l_status)
				across
					a_block.nodes as c
				until
					l_status.item /= 0
				loop
					c.item.accept (l_visitor)
				end

				inspect l_status.item
				when 0 then
					if a_block.out_edges.is_empty then
						Result := False
					else
						Result := across a_block.out_edges as c some
								needs_initialization_rec (a_node, c.item.to_block, a_visited)
							end
					end
				when 1 then
					Result := False
				when 2 then
					Result := True
				end
			else
				Result := False
			end
		end

invariant
	buffer /= Void
	possible_semicolon_positions /= Void

end
